import sys

import click

from mhd_model.commands.validate.announcement import validate_announcement_file_task
from mhd_model.commands.validate.mhd_file import validate_mhd_file_task


@click.group(name="validate", context_settings={"help_option_names": ["-h", "--help"]})
def validate_group():
    """utilities to validate MetabolomicsHub files."""
    pass


validate_group.add_command(validate_announcement_file_task)
validate_group.add_command(validate_mhd_file_task)


if __name__ == "__main__":
    if len(sys.argv) == 1:
        validate_group(["--help"])
    else:
        validate_group()



from __future__ import annotations

import json
import logging
import pathlib
from typing import Any, Generic, TypeVar
from urllib.parse import quote

import bioregistry
import httpx
from pydantic import BaseModel, ConfigDict, field_validator
from pydantic.alias_generators import to_camel, to_pascal

from mhd_model.model.v0_1.rules.cv_definitions import CONTROLLED_CV_DEFINITIONS
from mhd_model.shared.model import CvTerm
from mhd_model.shared.validation.definitions import ParentCvTerm

logger = logging.getLogger(__name__)

T = TypeVar("T")


class OlsBaseModel(BaseModel):
    """Base model class to convert python attributes to camel case"""

    model_config = ConfigDict(
        populate_by_name=True,
        alias_generator=to_camel,
        JSON_schema_serialization_defaults_required=True,
        field_title_generator=lambda field_name, field_info: to_pascal(
            field_name.replace("_", " ").strip()
        ),
    )


class OlsSearchModel(OlsBaseModel, Generic[T]):
    page: int
    num_elements: int
    total_pages: int
    total_elements: int
    elements: list[T]


class ChildrenSearchModel(OlsBaseModel):
    curie: str
    has_hierarchical_children: bool
    has_direct_children: bool
    iri: str
    is_obsolete: bool
    label: str
    ontology_preferred_prefix: str

    @field_validator("label", mode="before")
    @classmethod
    def label_validator(cls, value: list[str] | Any | None) -> str:
        if value is None:
            return ""
        if isinstance(value, list):
            return value[0] if value else ""
        return str(value)


class CvTermHelper:
    def __init__(self) -> None:
        self.cache: dict[str, None | dict[str, CvTerm]] = {}
        self.search_cache: dict[str, tuple[bool, str | None]] = {}

    def save_children(self, parent: ParentCvTerm, children: dict[str, CvTerm]) -> None:
        file_path = self.get_children_cache_file_path(parent)
        children_dict = {x: y.model_dump() for x, y in children.items()}
        with file_path.open("w") as f:
            f.write(
                json.dumps(
                    {
                        "parent": parent.model_dump(),
                        "children": children_dict,
                    },
                    indent=2,
                )
            )

    def get_children_cache_file_path(self, parent: ParentCvTerm) -> pathlib.Path:
        parent_path = pathlib.Path("cache")
        parent_path.mkdir(parents=True, exist_ok=True)
        parent_option = "p" if parent.allow_parent else "_"
        leaf_option = "l" if parent.allow_parent else "_"
        name_prefix = (
            parent.cv_term.accession.replace(":", "_") if parent.cv_term else ""
        )
        file_path = parent_path / pathlib.Path(
            f"{name_prefix}_children_{parent_option}_{leaf_option}.json"
        )
        return file_path

    def load_children(self, parent: ParentCvTerm) -> None | dict[str, CvTerm]:
        file_path = self.get_children_cache_file_path(parent)
        if file_path in self.cache:
            self.cache[file_path]
        if not file_path.exists():
            return None
        try:
            logger.debug(
                "Loading children CV Terms for %s - %s: %s",
                parent.cv_term.accession,
                parent.cv_term.name,
                file_path,
            )
            with file_path.open() as f:
                data = json.load(f)
            object_map = {
                x: CvTerm.model_validate(y) for x, y in data["children"].items()
            }
            # key = (parent.cv_term.source, parent.cv_term.accession)
            # if key not in self.cache:
            self.cache[file_path] = object_map
            logger.debug("Children CV Terms are loaded. %s", file_path)
            return object_map
        except Exception as ex:
            logger.exception(str(ex))
            return None

    def get_children_of_cv_term(self, parent: ParentCvTerm) -> dict[str, CvTerm]:
        file_path = self.get_children_cache_file_path(parent)
        if file_path in self.cache:
            return self.cache[file_path]

        children_map = self.load_children(parent)
        if children_map is not None:
            return children_map

        children: list[ChildrenSearchModel] = []
        if parent.allow_parent:
            uri = self.get_uri(parent.cv_term)
            children.append(
                ChildrenSearchModel(
                    curie=parent.cv_term.accession,
                    has_direct_children=True,
                    has_hierarchical_children=True,
                    iri=uri,
                    label=parent.cv_term.name,
                    ontology_preferred_prefix=parent.cv_term.source,
                    is_obsolete=False,
                )
            )
        excluded_cv_accessions = set()
        if parent.excluded_cv_terms:
            excluded_cv_accessions = {x.accession for x in parent.excluded_cv_terms}
        self.get_children(
            parent.cv_term,
            children,
            parent.allow_only_leaf,
            excluded_cv_accessions,
        )
        children.sort(key=lambda x: x.label)
        children_cv_terms = {
            x.curie: CvTerm(
                accession=x.curie, source=x.ontology_preferred_prefix, name=x.label
            )
            for x in children
        }
        self.cache[file_path] = children_cv_terms
        self.save_children(parent, children_cv_terms)
        return children_cv_terms

    def get_children(
        self,
        cv_term: CvTerm,
        children: list[ChildrenSearchModel],
        allow_only_leaf: bool = True,
        excluded_cv_accessions: None | set[str] = None,
    ) -> None:
        parent_uri = self.get_uri(cv_term)

        parent_uri_encoded = quote(quote(parent_uri, safe=[]))
        children_subpath = f"/ontologies/{cv_term.source.lower()}/classes/{parent_uri_encoded}/children"
        ols4_base_url = "https://www.ebi.ac.uk/ols4/api/v2"

        url = ols4_base_url + children_subpath
        page = 0
        finished = False
        headers = {"Accept": "application/json"}
        selected_terms = []
        while not finished:
            params = {"page": page, "size": 100}
            page += 1
            result = httpx.get(url, params=params, headers=headers, timeout=10)
            result_json = result.json()
            search = OlsSearchModel[ChildrenSearchModel].model_validate(result_json)
            if not excluded_cv_accessions:
                excluded_cv_accessions = set()
            selected = [
                x
                for x in search.elements
                if not x.is_obsolete and x.curie not in excluded_cv_accessions
            ]
            if selected:
                selected_terms.extend(selected)
            if page >= search.total_pages:
                finished = True
        for term in selected_terms:
            if not allow_only_leaf or (
                allow_only_leaf and not term.has_hierarchical_children
            ):
                children.append(term)

            if term.has_hierarchical_children:
                self.get_children(
                    cv_term=CvTerm(
                        accession=term.curie,
                        name=term.label,
                        source=term.ontology_preferred_prefix,
                    ),
                    children=children,
                    allow_only_leaf=allow_only_leaf,
                    excluded_cv_accessions=excluded_cv_accessions,
                )

    def get_uri_with_custom_convertor(self, cv_term: CvTerm) -> None | str:
        source = cv_term.source
        cv_definition = CONTROLLED_CV_DEFINITIONS.get(source)
        parent_uri = None
        accession = cv_term.source

        if cv_definition:
            accession = cv_term.accession
            if not accession:
                return ""
            if accession.startswith(cv_definition.prefix):
                parent_uri = accession
            elif ":" in accession:
                parent_uri = cv_definition.prefix + accession.split(":")[1]
            else:
                parent_uri = cv_definition.prefix + accession
        else:
            return None
        return parent_uri
        # search.page

    def get_uri(self, cv_term: CvTerm) -> str | None:
        uri = None
        if cv_term and ":" in cv_term.accession:
            uri = self.get_uri_with_custom_convertor(cv_term)
            if not uri:
                prefix, identifier = cv_term.accession.split(":")
                uri = bioregistry.get_default_iri(prefix, identifier)
                if uri and "https://www.ebi.ac.uk/ols/ontologies/edam/terms?" in uri:
                    uri = None

        return uri

    def check_cv_term(
        self, cv_term: CvTerm, parent_cv_term: None | ParentCvTerm = None
    ) -> tuple[bool, str]:
        if not cv_term.accession or not cv_term.name or not cv_term.source:
            message = f"Invalid cv term [{cv_term.source}, {cv_term.accession}, {cv_term.name}]"
            logger.error(message)
            return False, message

        parent = parent_cv_term.cv_term if parent_cv_term else None

        if parent:
            if not parent.accession or not parent.name or not parent.source:
                message = f"Invalid cv term parent [{parent.source}, {parent.accession}, {parent.name}"
                logger.error(message)
                return False, message

        key = ",".join([str(cv_term), str(parent)])

        if key in self.search_cache:
            return self.search_cache[key]

        children_subpath = "/search"
        if parent_cv_term:
            logger.debug(
                "Check CV term %s - %s whether is child of %s %s",
                cv_term.accession,
                cv_term.name,
                parent_cv_term.cv_term.accession,
                parent_cv_term.cv_term.name,
            )
        else:
            logger.debug("Check CV term %s - %s", cv_term.accession, cv_term.name)
        params = {
            "q": cv_term.accession,
            "ontology": cv_term.source,
            "type": "class,property",
            "queryFields": "obo_id",
            "fieldList": "iri,obo_id,label,short_form",
            "exact": True,
            "format": "json",
            "start": 0,
            "rows": 1,
            "local": False,
            "obsoletes": False,
            "lang": "en",
            # "isLeaf": (
            #     True if parent_cv_term and parent_cv_term.allow_only_leaf else False
            # ),
        }
        ols4_base_url = "https://www.ebi.ac.uk/ols4/api"
        url = ols4_base_url + children_subpath
        if parent_cv_term:
            parent_uri = self.get_uri(parent_cv_term.cv_term)
            params["allChildrenOf"] = parent_uri
            logger.info(
                "%s: %s in cv %s, parent %s",
                url,
                cv_term.accession,
                cv_term.source,
                parent_uri,
            )
        else:
            logger.info("%s: %s in cv %s", url, cv_term.accession, cv_term.source)

        headers = {"Accept": "application/json"}
        try:
            logger.debug("Searching %s", url)
            result = httpx.get(url, params=params, headers=headers, timeout=10)
            if result.status_code == 404:
                self.search_cache[key] = (
                    False,
                    f"{cv_term.source} is not valid or {cv_term.accession} is not in ontology {cv_term.source}",
                )
                return self.search_cache[key]
            result.raise_for_status()
            result_json = result.json()
            if result_json.get("response"):
                docs = result_json.get("response").get("docs")
                if docs and docs[0]["obo_id"] == cv_term.accession:
                    if parent_cv_term:
                        logger.debug(
                            "CV term %s - %s is child of %s %s",
                            cv_term.accession,
                            cv_term.name,
                            (
                                parent_cv_term.cv_term.accession
                                if parent_cv_term
                                else None
                            ),
                            parent_cv_term.cv_term.name if parent_cv_term else None,
                        )
                    else:
                        logger.debug(
                            "CV term %s - %s is valid CV term",
                            cv_term.accession,
                            cv_term.name,
                        )
                    self.search_cache[key] = (True, None)
                    return self.search_cache[key]
                if not parent_cv_term:
                    self.search_cache[key] = (
                        False,
                        f"'{cv_term.source}' is not valid source or [{cv_term.source}, {cv_term.accession}, {cv_term.name}] is not found in {cv_term.source} ontology",
                    )
                else:
                    self.search_cache[key] = (
                        False,
                        f"{cv_term.name} {cv_term.accession} is not child of {parent.accession} on source {parent.source}. ",
                    )
                return self.search_cache[key]

            else:
                self.search_cache[key] = (False, f"{cv_term.accession} does not match")
                return self.search_cache[key]

        except httpx.HTTPStatusError as ex:
            return False, f"{cv_term.accession} search failed: {str(ex)}"
        except Exception as ex:
            return (
                False,
                f"{cv_term.accession} is not in {cv_term.source} ontology. {str(ex)}",
            )



from mhd_model.shared.model import CvTerm

MANAGED_CV_TERM_OBJECTS = {
    "characteristic-type",
    "characteristic-value",
    "metabolite-identifier",
    "data-provider",
    "descriptor",
    "factor-type",
    "factor-value",
    "parameter-type",
    "parameter-value",
    "protocol-type",
}

COMMON_TECHNOLOGY_TYPES = {
    "OBI:0000470": CvTerm(
        source="OBI",
        accession="OBI:0000470",
        name="mass spectrometry assay",
    ),
    "OBI:0000623": CvTerm(
        source="OBI",
        accession="OBI:0000623",
        name="NMR spectroscopy assay",
    ),
}

COMMON_ASSAY_TYPES = {
    "OBI:0003097S": CvTerm(
        source="OBI",
        accession="OBI:0003097",
        name="liquid chromatography mass spectrometry assay",
    ),
    "OBI:0003110": CvTerm(
        source="OBI",
        accession="OBI:0003110",
        name="gas chromatography mass spectrometry assay",
    ),
    "OBI:0000470": CvTerm(
        source="OBI",
        accession="OBI:0000470",
        name="mass spectrometry assay",
    ),
    "OBI:0000623": CvTerm(
        source="OBI",
        accession="OBI:0000623",
        name="NMR spectroscopy assay",
    ),
    # CE-MS
    # DI-MS
    # FIA-MS
    # GCxGC-MS
    # MALDI-MS
    # MSImaging
    # GC-FID
    # LC-DAD
    # MRImaging
}

COMMON_MEASUREMENT_TYPES = {
    "MSIO:0000100": CvTerm(
        source="MSIO",
        accession="MSIO:0000100",
        name="targeted metabolite profiling",
    ),
    "MSIO:0000101": CvTerm(
        source="MSIO",
        accession="MSIO:0000101",
        name="untargeted metabolite profiling",
    ),
    # TODO others?
}

COMMON_OMICS_TYPES = {
    "EDAM:3172": CvTerm(
        source="EDAM",
        accession="EDAM:3172",
        name="Metabolomics",
    ),
    "EDAM:0153": CvTerm(
        source="EDAM",
        accession="EDAM:0153",
        name="Lipidomics",
    ),
    "EDAM:3955": CvTerm(
        source="EDAM",
        accession="EDAM:3955",
        name="Fluxomics",
    ),
    # TODO others?
}

MISSING_PUBLICATION_REASON = {
    "MS:1002853": CvTerm(
        source="MS",
        accession="MS:1002853",
        name="no associated published manuscript",
    ),
    "MS:1002858": CvTerm(
        source="MS",
        accession="MS:1002858",
        name="Dataset with its publication pending",
    ),
}

COMMON_CHARACTERISTIC_DEFINITIONS = {
    "NCIT:C14250": CvTerm(source="NCIT", accession="NCIT:C14250", name="organism"),
    "NCIT:C103199": CvTerm(
        source="NCIT", accession="NCIT:C103199", name="organism part"
    ),
    "EFO:0000408": CvTerm(source="EFO", accession="EFO:0000408", name="disease"),
    "EFO:0000324": CvTerm(source="EFO", accession="EFO:0000324", name="cell type"),
    # "GAZ:00000448": CvTerm(
    #     source="EFO", accession="GAZ:00000448", name="geographical location"
    # ),
}


REQUIRED_CHARACTERISTIC_DEFINITIONS = {
    "NCIT:C14250": COMMON_CHARACTERISTIC_DEFINITIONS["NCIT:C14250"],
    "NCIT:C103199": COMMON_CHARACTERISTIC_DEFINITIONS["NCIT:C103199"],
}

COMMON_STUDY_FACTOR_DEFINITIONS = {
    "EFO:0000408": CvTerm(source="EFO", accession="EFO:0000408", name="disease"),
}


REQUIRED_COMMON_PARAMETER_DEFINITIONS = {
    "MSIO:0000171": CvTerm(
        source="MSIO", accession="MSIO:0000171", name="mass spectrometry instrument"
    ),
    "OBI:0000485": CvTerm(
        source="OBI", accession="OBI:0000485", name="chromatography instrument"
    ),
}

COMMON_PARAMETER_DEFINITIONS = REQUIRED_COMMON_PARAMETER_DEFINITIONS.copy()
COMMON_PARAMETER_DEFINITIONS.update(
    {
        "MS:1000465": CvTerm(source="MS", accession="MS:1000465", name="scan polarity"),
        "OBI:0000521": CvTerm(
            source="OBI", accession="OBI:0000521", name="flame ionization detector"
        ),
        "CHMO:0002503": CvTerm(
            source="CHMO", accession="CHMO:0002503", name="diode array detector"
        ),
        "OBI:0001132": CvTerm(
            source="OBI",
            accession="OBI:0001132",
            name="capillary electrophoresis instrument",
        ),
        "OBI:0000345": CvTerm(
            source="OBI", accession="OBI:0000345", name="mass analyzer"
        ),
        "CHMO:0000960": CvTerm(
            source="CHMO", accession="CHMO:0000960", name="ion source"
        ),
        "MTBLS:50001": CvTerm(
            source="MTBLS",
            accession="MTBLS:50001",
            name="column model",
        ),
        "MTBLS:50002": CvTerm(
            source="MTBLS",
            accession="MTBLS:50002",
            name="column type",
        ),
        "MTBLS:50003": CvTerm(
            source="MTBLS",
            accession="MTBLS:50003",
            name="guard column",
        ),
        "MTBLS:50004": CvTerm(
            source="MTBLS",
            accession="MTBLS:50004",
            name="autosampler model",
        ),
        "MTBLS:50010": CvTerm(
            source="MTBLS",
            accession="MTBLS:50010",
            name="post extraction",
        ),
        "MTBLS:50011": CvTerm(
            source="MTBLS",
            accession="MTBLS:50011",
            name="derivatization",
        ),
        "MTBLS:50020": CvTerm(
            source="MTBLS",
            accession="MTBLS:50020",
            name="scan m/z range",
        ),
        "MTBLS:50021": CvTerm(
            source="MTBLS",
            accession="MTBLS:50021",
            name="fia instrument",
        ),
        # column model
        # column type
        # scan m/z range
        # flow injection analysis (FIA) instrument
    }
)

REQUIRED_PARAMETER_DEFINITIONS = {
    "LC-MS": {
        "CHMO:0000470": {
            "MSIO:0000171": REQUIRED_COMMON_PARAMETER_DEFINITIONS["MSIO:0000171"],
        }
    },
    "GC-MS": {
        "CHMO:0000470": {
            "MSIO:0000171": REQUIRED_COMMON_PARAMETER_DEFINITIONS["MSIO:0000171"],
        }
    },
    "CE-MS": {
        "CHMO:0000470": {
            "MSIO:0000171": REQUIRED_COMMON_PARAMETER_DEFINITIONS["MSIO:0000171"],
        }
    },
    "DI-MS": {
        "CHMO:0000470": {
            "MSIO:0000171": REQUIRED_COMMON_PARAMETER_DEFINITIONS["MSIO:0000171"],
        }
    },
    "FIA-MS": {
        "CHMO:0000470": {
            "MSIO:0000171": REQUIRED_COMMON_PARAMETER_DEFINITIONS["MSIO:0000171"],
        }
    },
    "GCxGC-MS": {
        "CHMO:0000470": {
            "MSIO:0000171": REQUIRED_COMMON_PARAMETER_DEFINITIONS["MSIO:0000171"],
        }
    },
    ####### OTHERS ######
    # MR Imaging, LC-DAD, GC-FID, NMR
}

COMMON_PROTOCOL_PARAMETERS = {
    "CHMO:0000470": {
        "MS:1000465": COMMON_PARAMETER_DEFINITIONS["MS:1000465"],
        "MTBLS:50020": COMMON_PARAMETER_DEFINITIONS["MTBLS:50020"],
        "MSIO:0000171": COMMON_PARAMETER_DEFINITIONS["MSIO:0000171"],
        "CHMO:0000960": COMMON_PARAMETER_DEFINITIONS["CHMO:0000960"],
        "OBI:0000345": COMMON_PARAMETER_DEFINITIONS["OBI:0000345"],
    },
    "CHMO:0001000": {
        "OBI:0000485": COMMON_PARAMETER_DEFINITIONS["OBI:0000485"],
        "MTBLS:50001": COMMON_PARAMETER_DEFINITIONS["MTBLS:50001"],
        "MTBLS:50002": COMMON_PARAMETER_DEFINITIONS["MTBLS:50002"],
        "MTBLS:50003": COMMON_PARAMETER_DEFINITIONS["MTBLS:50003"],
        "MTBLS:50004": COMMON_PARAMETER_DEFINITIONS["MTBLS:50004"],
    },
}

COMMON_PROTOCOLS = {
    "EFO:0005518": CvTerm(
        source="EFO",
        accession="EFO:0005518",
        name="sample collection protocol",
    ),
    "MS:1000831": CvTerm(
        source="MS",
        accession="MS:1000831",
        name="sample preparation",
    ),
    "CHMO:0000470": CvTerm(
        source="CHMO",
        accession="CHMO:0000470",
        name="mass spectrometry",
    ),
    "OBI:0200000": CvTerm(
        source="OBI",
        accession="OBI:0200000",
        name="data transform",
    ),
    "MI:2131": CvTerm(
        source="MI",
        accession="MI:2131",
        name="metabolite identification",
    ),
    "CHMO:0001000": CvTerm(
        source="CHMO",
        accession="CHMO:0001000",
        name="chromatography",
    ),
    "EFO:0003969": CvTerm(
        source="EFO",
        accession="EFO:0003969",
        name="treatment protocol",
    ),
    "CHMO:0001024": CvTerm(
        source="CHMO",
        accession="CHMO:0001024",
        name="capillary electrophoresis",
    ),
    "MS:1000058": CvTerm(
        source="MS",
        accession="MS:1000058",
        name="flow injection analysis",
    ),
    ###### DI-MS ######
    # Direct infusion
    ###### SPE-IMS-MS ######
    # Solid-Phase Extraction Ion Mobility Spectrometry
    ###### MSImaging ######
    # Preparation
    # Histology
    ###### MRImaging ######
    # Magnetic resonance imaging
    # In vivo magnetic resonance spectroscopy
    # In vivo magnetic resonance assay
    ###### NMR ######
    # NMR sample
    # NMR spectroscopy
    # NMR assay
}

# TODO Add all of them
assay_technique_protocols = {
    "CE-MS": [
        "Capillary Electrophoresis",
        "Mass spectrometry",
    ],
    "DI-MS": [
        "Direct infusion",
        "Mass spectrometry",
    ],
    "FIA-MS": [
        "Flow Injection Analysis",
        "Mass spectrometry",
    ],
    "GC-FID": [
        "Chromatography",
    ],
    "GC-MS": [
        "Chromatography",
        "Mass spectrometry",
    ],
    "GCxGC-MS": [
        "Sample collection",
        "Extraction",
        "Chromatography",
        "Mass spectrometry",
    ],
    "LC-DAD": [
        "Sample collection",
        "Extraction",
        "Chromatography",
        "Data transformation",
        "Metabolite identification",
    ],
    "LC-MS": [
        "Sample collection",
        "Extraction",
        "Chromatography",
        "Mass spectrometry",
        "Data transformation",
        "Metabolite identification",
    ],
    "MALDI-MS": [
        "Sample collection",
        "Extraction",
        "Mass spectrometry",
        "Data transformation",
        "Metabolite identification",
    ],
    "MRImaging": [
        "Sample collection",
        "Magnetic resonance imaging",
        "In vivo magnetic resonance spectroscopy",
        "In vivo magnetic resonance assay",
        "Data transformation",
        "Metabolite identification",
    ],
    "MS": [
        "Sample collection",
        "Extraction",
        "Mass spectrometry",
        "Data transformation",
        "Metabolite identification",
    ],
    "MSImaging": [
        "Sample collection",
        "Preparation",
        "Mass spectrometry",
        "Histology",
        "Data transformation",
        "Metabolite identification",
    ],
    "NMR": [
        "Sample collection",
        "Extraction",
        "NMR sample",
        "NMR spectroscopy",
        "NMR assay",
        "Data transformation",
        "Metabolite identification",
    ],
    "SPE-IMS-MS": [
        "Sample collection",
        "Extraction",
        "Solid-Phase Extraction Ion Mobility Spectrometry",
        "Mass spectrometry",
        "Data transformation",
        "Metabolite identification",
    ],
    "TD-GC-MS": [
        "Sample collection",
        "Extraction",
        "Chromatography",
        "Mass spectrometry",
        "Data transformation",
        "Metabolite identification",
    ],
}



import sys

import click

from mhd_model import __version__
from mhd_model.commands.create.create import create_group
from mhd_model.commands.validate.validate import validate_group


@click.group(context_settings={"help_option_names": ["-h", "--help"]})
@click.version_option(__version__)
def cli():
    """MetabomicsHub CLI with subcommands."""
    pass


cli.add_command(create_group)
cli.add_command(validate_group)

if __name__ == "__main__":
    if len(sys.argv) == 1:
        cli(["--help"])
    else:
        cli()



from typing import Annotated

from pydantic import Field

from mhd_model.model.v0_1.announcement.profiles.base.fields import (
    ExtendedCvTermKeyValue,
)
from mhd_model.model.v0_1.announcement.profiles.base.profile import AnnouncementProtocol
from mhd_model.model.v0_1.announcement.validation.definitions import (
    CheckChildCvTermKeyValues,
    CheckCvTermKeyValue,
    CheckCvTermKeyValues,
)
from mhd_model.model.v0_1.rules.managed_cv_terms import (
    COMMON_ASSAY_TYPES,
    COMMON_PROTOCOLS,
    COMMON_TECHNOLOGY_TYPES,
    REQUIRED_COMMON_PARAMETER_DEFINITIONS,
)
from mhd_model.shared.model import CvTerm, CvTermKeyValue, CvTermValue
from mhd_model.shared.validation.definitions import (
    AllowedChildrenCvTerms,
    AllowedCvList,
    AllowedCvTerms,
    ParentCvTerm,
)

MetaboliteDatabaseId = Annotated[
    CvTermValue,
    Field(
        json_schema_extra={
            "profileValidation": AllowedChildrenCvTerms(
                parent_cv_terms=[
                    ParentCvTerm(
                        cv_term=CvTerm(
                            accession="CHEMINF:000464",
                            source="CHEMINF",
                            name="chemical database identifier",
                        ),
                        allow_only_leaf=False,
                        index_cv_terms=False,
                    )
                ],
                allowed_other_sources=["REFMET"],
            ).model_dump(by_alias=True)
        }
    ),
]

MsTechnologyType = Annotated[
    CvTerm,
    Field(
        json_schema_extra={
            "profileValidation": AllowedCvTerms(
                cv_terms=[COMMON_TECHNOLOGY_TYPES["OBI:0000470"]]
            ).model_dump(by_alias=True)
        },
    ),
]

ExtendedCharacteristicValues = Annotated[
    list[CvTermKeyValue],
    Field(
        min_length=1,
        json_schema_extra={
            "profileValidation": CheckCvTermKeyValues(
                required_items=[
                    CheckCvTermKeyValue(
                        cv_term_key=CvTerm(
                            source="NCIT", accession="NCIT:C14250", name="organism"
                        ),
                        controls=[
                            AllowedCvList(
                                source_names=["ENVO", "NCBITAXON"],
                                allowed_other_sources=["wikidata", "ILX"],
                            )
                        ],
                        min_value_count=1,
                    ),
                    CheckCvTermKeyValue(
                        cv_term_key=CvTerm(
                            source="NCIT",
                            accession="NCIT:C103199",
                            name="organism part",
                        ),
                        controls=[
                            AllowedCvList(
                                source_names=["UBERON", "BTO", "NCIT", "SNOMED", "MSIO"]
                            )
                        ],
                        min_value_count=1,
                    ),
                ],
                optional_items=[
                    CheckCvTermKeyValue(
                        cv_term_key=CvTerm(
                            source="EFO", accession="EFO:0000408", name="disease"
                        ),
                        controls=[
                            AllowedCvList(
                                source_names=["DOID", "HP", "MP", "SNOMED"],
                                allowed_other_sources=["wikidata", "ILX"],
                            )
                        ],
                        min_value_count=1,
                    ),
                    CheckCvTermKeyValue(
                        cv_term_key=CvTerm(
                            source="EFO", accession="EFO:0000324", name="cell type"
                        ),
                        controls=[AllowedCvList(source_names=["CL", "CLO"])],
                        min_value_count=1,
                    ),
                ],
            ).model_dump(serialize_as_any=True, by_alias=True)
        },
    ),
]

MsAssayType = Annotated[
    CvTerm,
    Field(
        json_schema_extra={
            "profileValidation": AllowedCvTerms(
                cv_terms=list(COMMON_ASSAY_TYPES.values())
            ).model_dump(by_alias=True)
        },
    ),
]

StudyFactors = Annotated[
    list[ExtendedCvTermKeyValue],
    Field(
        min_length=1,
        json_schema_extra={
            "profileValidation": CheckCvTermKeyValues(
                optional_items=[
                    CheckCvTermKeyValue(
                        cv_term_key=CvTerm(
                            source="EFO", accession="EFO:0000408", name="disease"
                        ),
                        controls=[
                            AllowedCvList(
                                source_names=["DOID", "HP", "MP", "SNOMED"],
                                allowed_other_sources=["wikidata", "ILX"],
                            )
                        ],
                        min_value_count=1,
                    )
                ]
            ).model_dump(serialize_as_any=True, by_alias=True)
        },
    ),
]

Protocols = Annotated[
    list[AnnouncementProtocol],
    Field(
        json_schema_extra={
            "profileValidation": CheckChildCvTermKeyValues(
                conditional_field_name="protocol_type",
                conditional_cv_term=COMMON_PROTOCOLS["CHMO:0000470"],
                key_values_field_name="protocol_parameters",
                key_values_control=CheckCvTermKeyValues(
                    required_items=[
                        CheckCvTermKeyValue(
                            cv_term_key=REQUIRED_COMMON_PARAMETER_DEFINITIONS[
                                "MSIO:0000171"
                            ],
                            controls=[
                                AllowedChildrenCvTerms(
                                    parent_cv_terms=[
                                        ParentCvTerm(
                                            cv_term=CvTerm(
                                                source="MS",
                                                accession="MS:1000031",
                                                name="instrument model",
                                            ),
                                            allow_only_leaf=True,
                                        ),
                                    ],
                                )
                            ],
                        )
                    ]
                ),
            ).model_dump(serialize_as_any=True, by_alias=True)
        }
    ),
]



from typing import Annotated

from pydantic import BaseModel, Field


class SupportedJsonSchema(BaseModel):
    uri: str
    file_path: Annotated[str, Field(exclude=True)]


class SupportedSchema(SupportedJsonSchema):
    default_profile_uri: str
    supported_profiles: dict[str, SupportedJsonSchema]


class SupportedSchemaMap(BaseModel):
    default_schema_uri: str
    schemas: dict[str, SupportedSchema]


ANNOUNCEMENT_FILE_V0_1_DEFAULT_SCHEMA_NAME = "https://metabolomicshub.github.io/mhd-model/schemas/v0_1/announcement-v0.1.schema.json"
ANNOUNCEMENT_FILE_V0_1_MS_PROFILE_NAME = "https://metabolomicshub.github.io/mhd-model/schemas/v0_1/announcement-v0.1.schema.ms-profile.json"
ANNOUNCEMENT_FILE_V0_1_LEGACY_PROFILE_NAME = "https://metabolomicshub.github.io/mhd-model/schemas/v0_1/announcement-v0.1.schema.legacy-profile.json"

MHD_MODEL_V0_1_DEFAULT_SCHEMA_NAME = "https://metabolomicshub.github.io/mhd-model/schemas/v0_1/common-data-model-v0.1.schema.json"
MHD_MODEL_V0_1_MS_PROFILE_NAME = "https://metabolomicshub.github.io/mhd-model/schemas/v0_1/common-data-model-v0.1.ms-profile.json"
MHD_MODEL_V0_1_LEGACY_PROFILE_NAME = "https://metabolomicshub.github.io/mhd-model/schemas/v0_1/common-data-model-v0.1.legacy-profile.json"

SUPPORTED_SCHEMA_MAP = SupportedSchemaMap(
    default_schema_uri=ANNOUNCEMENT_FILE_V0_1_DEFAULT_SCHEMA_NAME,
    schemas={
        ANNOUNCEMENT_FILE_V0_1_DEFAULT_SCHEMA_NAME: SupportedSchema(
            uri=ANNOUNCEMENT_FILE_V0_1_DEFAULT_SCHEMA_NAME,
            file_path="mhd_model/schemas/mhd/announcement-v0.1.schema.json",
            default_profile_uri=ANNOUNCEMENT_FILE_V0_1_MS_PROFILE_NAME,
            supported_profiles={
                ANNOUNCEMENT_FILE_V0_1_MS_PROFILE_NAME: SupportedJsonSchema(
                    uri=ANNOUNCEMENT_FILE_V0_1_MS_PROFILE_NAME,
                    file_path="mhd_model/schemas/mhd/announcement-v0.1.schema.ms-profile.json",
                ),
                ANNOUNCEMENT_FILE_V0_1_LEGACY_PROFILE_NAME: SupportedJsonSchema(
                    uri=ANNOUNCEMENT_FILE_V0_1_LEGACY_PROFILE_NAME,
                    file_path="mhd_model/schemas/mhd/announcement-v0.1.schema.legacy-profile.json",
                ),
            },
        ),
        MHD_MODEL_V0_1_DEFAULT_SCHEMA_NAME: SupportedSchema(
            uri=MHD_MODEL_V0_1_DEFAULT_SCHEMA_NAME,
            file_path="mhd_model/schemas/mhd/common-data-model-v0.1.schema.json",
            default_profile_uri=MHD_MODEL_V0_1_MS_PROFILE_NAME,
            supported_profiles={
                MHD_MODEL_V0_1_MS_PROFILE_NAME: SupportedJsonSchema(
                    uri=MHD_MODEL_V0_1_MS_PROFILE_NAME,
                    file_path="mhd_model/schemas/mhd/common-data-model-v0.1.ms-profile.json",
                ),
                MHD_MODEL_V0_1_LEGACY_PROFILE_NAME: SupportedJsonSchema(
                    uri=MHD_MODEL_V0_1_LEGACY_PROFILE_NAME,
                    file_path="mhd_model/schemas/mhd/common-data-model-v0.1.legacy-profile.json",
                ),
            },
        ),
    },
)



from typing import Annotated

from pydantic import Field

MhdIdentifier = Annotated[
    str, Field(pattern=r"^MHD[A-Z][0-9]{6,8}$", title="MHD Identifier")
]

PubMedId = Annotated[str, Field(pattern=r"^[0-9]{1,20}$", title="PubMed Id")]
ORCID = Annotated[
    str, Field(pattern=r"^[0-9]{4}-[0-9]{4}-[0-9]{4}-[0-9]{3}[X0-9]$", title="ORCID")
]
DOI = Annotated[str, Field(pattern=r"^10[.].+/.+$", title="DOI")]


Author = Annotated[str, Field(min_length=2, title="Author")]
Authors = Annotated[list[Author], Field(min_length=1, title="Authors")]
GrantId = Annotated[str, Field(min_length=2, title="Grant Identifier")]



import datetime
import decimal

from pydantic import BaseModel, ConfigDict, Field
from pydantic.alias_generators import to_pascal
from typing_extensions import Annotated


class MhdConfigModel(BaseModel):
    """Base model class to convert python attributes to camel case"""

    model_config = ConfigDict(
        populate_by_name=True,
        JSON_schema_serialization_defaults_required=True,
        field_title_generator=lambda field_name, field_info: to_pascal(
            field_name.replace("_", " ").strip()
        ),
        # alias_generator=to_camel,
    )


class CvTerm(MhdConfigModel):
    source: Annotated[
        str,
        Field(description="Ontology source name."),
    ] = ""
    accession: Annotated[
        str,
        Field(description="Accession number of CV term in compact URI format."),
    ] = ""
    name: Annotated[
        str,
        Field(description="Label of CV term."),
    ] = ""

    def get_unique_id(self) -> str:
        return f"{self.source or ''},{self.accession or ''},{self.name or ''}"

    def __hash__(self) -> int:
        return hash(self.get_unique_id())

    def __lt__(self, other: "CvTerm") -> bool:
        if isinstance(other, CvTerm):
            return self.get_unique_id() < other.get_unique_id()
        return NotImplemented

    def get_label(self) -> str:
        return f"[{self.source or ''}, {self.accession or ''}, {self.name or ''}]"

    def __str__(self) -> str:
        return self.get_label()


class UnitCvTerm(CvTerm): ...


class QuantitativeValue(MhdConfigModel):
    value: None | str | int | float | decimal.Decimal = None
    unit: None | UnitCvTerm = None


class CvTermValue(CvTerm, QuantitativeValue):
    def get_unique_id(self) -> str:
        unit_key = self.unit.get_unique_id() if self.unit else ""
        value_key = (
            self.value.get_unique_id()
            if isinstance(self.value, CvTerm)
            else str(self.value)
            if self.value is not None
            else ""
        )

        return f"{super().get_unique_id()},{value_key or ''},{unit_key or ''}"

    def get_label(self) -> str:
        unit_key = self.unit.get_label() if self.unit else ""
        value_key = (
            self.value.get_label()
            if isinstance(self.unit, CvTerm) and self.value
            else str(self.value)
            if self.value is not None
            else ""
        )

        return f"[{self.source or ''}, {self.accession or ''}, {self.name or ''}, {value_key or ''}, {unit_key or ''}]"


class CvTermKeyValue(MhdConfigModel):
    key: Annotated[CvTerm, Field()]
    values: Annotated[None | list[CvTerm] | list[QuantitativeValue], Field()] = None


class CvDefinition(MhdConfigModel):
    label: str = ""
    name: str = ""
    uri: str = ""
    prefix: str = ""
    alternative_labels: Annotated[None | list[str], Field(exclude=True)] = None
    alternative_prefixes: Annotated[None | list[str], Field(exclude=True)] = None


class Revision(MhdConfigModel):
    revision: Annotated[int, Field()]
    revision_datetime: datetime.datetime
    comment: str


class BaseMhdDataset(MhdConfigModel):
    repository_name: Annotated[None | str, Field()] = None
    mhd_identifier: Annotated[None | str, Field()] = None
    repository_identifier: Annotated[None | str, Field()] = None
    revision: Annotated[None | int, Field()] = None
    revision_datetime: Annotated[None | datetime.datetime, Field()] = None
    repository_revision: Annotated[None | int, Field()] = None
    repository_revision_datetime: Annotated[None | datetime.datetime, Field()] = None

    change_log: Annotated[
        None | list[Revision], Field(min_length=1, description="Revision")
    ] = None


class ProfileEnabledDataset(BaseMhdDataset):
    schema_name: Annotated[
        str, Field(alias="$schema", description="Schema name of the file")
    ]
    profile_uri: Annotated[str, Field(description="Validation Profiles")]


class CvEnabledDataset(ProfileEnabledDataset):
    cv_definitions: Annotated[list[CvDefinition], Field()] = []



from typing import Annotated

from pydantic import Field

from mhd_model.model.v0_1.rules.managed_cv_terms import (
    COMMON_ASSAY_TYPES,
    COMMON_MEASUREMENT_TYPES,
    COMMON_OMICS_TYPES,
    COMMON_PROTOCOLS,
    COMMON_TECHNOLOGY_TYPES,
    MISSING_PUBLICATION_REASON,
)
from mhd_model.shared.model import (
    CvTerm,
    CvTermKeyValue,
    CvTermValue,
)
from mhd_model.shared.validation.definitions import (
    AccessibleCompactURI,
    AllowAnyCvTerm,
    AllowedChildrenCvTerms,
    AllowedCvTerms,
    CvTermPlaceholder,
    ParentCvTerm,
)

DOI = Annotated[
    str,
    Field(
        pattern=r"^10[.].+/.+$",
        json_schema_extra={
            "profileValidation": AccessibleCompactURI(default_prefix="doi").model_dump(
                by_alias=True
            )
        },
    ),
]


ORCID = Annotated[
    str,
    Field(
        pattern=r"^[0-9]{4}-[0-9]{4}-[0-9]{4}-[0-9]{3}[X0-9]$",
        json_schema_extra={
            "profileValidation": AccessibleCompactURI(
                default_prefix="orcid"
            ).model_dump(by_alias=True)
        },
    ),
]

PubMedId = Annotated[
    str,
    Field(
        pattern=r"^[0-9]{1,20}$",
        title="PubMed Id",
    ),
]


CvTermOrStr = Annotated[
    CvTerm,
    Field(
        json_schema_extra={
            "profileValidation": AllowAnyCvTerm(
                allowed_placeholder_values=[CvTermPlaceholder()],
            ).model_dump(by_alias=True)
        }
    ),
]

MissingPublicationReason = Annotated[
    CvTerm,
    Field(
        json_schema_extra={
            "profileValidation": AllowedCvTerms(
                cv_terms=list(MISSING_PUBLICATION_REASON.values())
            ).model_dump(by_alias=True)
        },
    ),
]

MetaboliteDatabaseId = Annotated[
    CvTermValue,
    Field(
        json_schema_extra={
            "profileValidation": AllowedChildrenCvTerms(
                parent_cv_terms=[
                    ParentCvTerm(
                        cv_term=CvTerm(
                            accession="CHEMINF:000464",
                            source="CHEMINF",
                            name="chemical database identifier",
                        ),
                        allow_only_leaf=False,
                        index_cv_terms=False,
                    )
                ],
                allowed_other_sources=["REFMET"],
                allowed_placeholder_values=[CvTermPlaceholder()],
            ).model_dump(by_alias=True)
        }
    ),
]

FactorDefinition = Annotated[
    CvTerm,
    Field(
        json_schema_extra={
            "profileValidation": AllowAnyCvTerm(
                allowed_placeholder_values=[CvTermPlaceholder()],
            ).model_dump(by_alias=True)
        }
    ),
]


ProtocolType = Annotated[
    CvTerm,
    Field(
        json_schema_extra={
            "profileValidation": AllowedCvTerms(
                cv_terms=list(COMMON_PROTOCOLS.values()),
                allowed_placeholder_values=[CvTermPlaceholder()],
            ).model_dump(by_alias=True)
        }
    ),
]


MeasurementType = Annotated[
    CvTerm,
    Field(
        json_schema_extra={
            "profileValidation": AllowedCvTerms(
                cv_terms=list(COMMON_MEASUREMENT_TYPES.values()),
                allowed_placeholder_values=[CvTermPlaceholder()],
            ).model_dump(by_alias=True)
        },
    ),
]


OmicsType = Annotated[
    CvTerm,
    Field(
        json_schema_extra={
            "profileValidation": AllowedCvTerms(
                cv_terms=list(COMMON_OMICS_TYPES.values()),
                allowed_placeholder_values=[CvTermPlaceholder()],
            ).model_dump(by_alias=True)
        },
    ),
]

TechnologyType = Annotated[
    CvTerm,
    Field(
        json_schema_extra={
            "profileValidation": AllowedCvTerms(
                cv_terms=list(COMMON_TECHNOLOGY_TYPES.values()),
                allowed_placeholder_values=[CvTermPlaceholder()],
            ).model_dump(by_alias=True)
        },
    ),
]


AssayType = Annotated[
    CvTerm,
    Field(
        json_schema_extra={
            "profileValidation": AllowedCvTerms(
                cv_terms=list(COMMON_ASSAY_TYPES.values()),
                allowed_placeholder_values=[CvTermPlaceholder()],
            ).model_dump(by_alias=True)
        },
    ),
]


class ExtendedCvTermKeyValue(CvTermKeyValue):
    key: Annotated[
        CvTerm,
        Field(
            json_schema_extra={
                "profileValidation": AllowAnyCvTerm(
                    allowed_placeholder_values=[CvTermPlaceholder()],
                ).model_dump(by_alias=True)
            }
        ),
    ]


RawDataFileFormat = Annotated[
    CvTerm,
    Field(
        json_schema_extra={
            "profileValidation": AllowedChildrenCvTerms(
                parent_cv_terms=[
                    ParentCvTerm(
                        cv_term=CvTerm(
                            source="EDAM",
                            accession="EDAM:1915",
                            name="Format",
                        ),
                        index_cv_terms=False,
                    ),
                    ParentCvTerm(
                        cv_term=CvTerm(
                            source="MS",
                            accession="MS:1000560",
                            name="mass spectrometer file format",
                        ),
                    ),
                ],
            ).model_dump(by_alias=True)
        }
    ),
]


CompressionFormat = Annotated[
    CvTerm,
    Field(
        json_schema_extra={
            "profileValidation": AllowedChildrenCvTerms(
                parent_cv_terms=[
                    ParentCvTerm(
                        cv_term=CvTerm(
                            source="EDAM",
                            accession="EDAM:1915",
                            name="Format",
                        ),
                        index_cv_terms=False,
                    )
                ]
            ).model_dump(by_alias=True)
        }
    ),
]

MetadataFileFormat = Annotated[
    CvTerm,
    Field(
        json_schema_extra={
            "profileValidation": AllowedChildrenCvTerms(
                parent_cv_terms=[
                    ParentCvTerm(
                        cv_term=CvTerm(
                            source="EDAM",
                            accession="EDAM:1915",
                            name="Format",
                        ),
                        index_cv_terms=False,
                    )
                ],
                allowed_placeholder_values=[CvTermPlaceholder()],
            ).model_dump(by_alias=True)
        }
    ),
]

ResultFileFormat = Annotated[
    CvTerm,
    Field(
        json_schema_extra={
            "profileValidation": AllowedChildrenCvTerms(
                parent_cv_terms=[
                    ParentCvTerm(
                        cv_term=CvTerm(
                            source="EDAM",
                            accession="EDAM:1915",
                            name="Format",
                        ),
                        index_cv_terms=False,
                    )
                ]
            ).model_dump(by_alias=True)
        }
    ),
]

DerivedFileFormat = Annotated[
    CvTerm,
    Field(
        json_schema_extra={
            "profileValidation": AllowedChildrenCvTerms(
                parent_cv_terms=[
                    ParentCvTerm(
                        cv_term=CvTerm(
                            source="EDAM",
                            accession="EDAM:1915",
                            name="Format",
                        ),
                        index_cv_terms=False,
                        allow_only_leaf=False,
                    ),
                ]
            ).model_dump(by_alias=True)
        }
    ),
]

SupplementaryFileFormat = Annotated[
    CvTerm,
    Field(
        json_schema_extra={
            "profileValidation": AllowedChildrenCvTerms(
                parent_cv_terms=[
                    ParentCvTerm(
                        cv_term=CvTerm(
                            source="EDAM",
                            accession="EDAM:1915",
                            name="Format",
                        ),
                        index_cv_terms=False,
                    )
                ]
            ).model_dump(by_alias=True)
        }
    ),
]



from __future__ import annotations

import logging
from ftplib import FTP
from typing import Any

import jsonschema
import reachable
from jsonschema import protocols, validators
from pydantic import ValidationError

from mhd_model.model.definitions import (
    SUPPORTED_SCHEMA_MAP,
)
from mhd_model.model.v0_1.announcement.validation.definitions import (
    CheckChildCvTermKeyValues,
    CheckCvTermKeyValue,
    CheckCvTermKeyValues,
    ProfileValidation,
)
from mhd_model.schema_utils import load_mhd_json_schema
from mhd_model.shared.model import CvTerm, CvTermKeyValue
from mhd_model.shared.validation.cv_term_helper import (
    CvTermHelper,
)
from mhd_model.shared.validation.validators import (
    AccessibleCompactURIValidator,
    AccessibleURIValidator,
    AllowAnyCvTermValidator,
    AllowedChildrenCvTermsValidator,
    AllowedCvListValidator,
    AllowedCvTermValidator,
    BaseProfileValidator,
    ProfileValidationGroupValidator,
    ValidationResult,
)

logger = logging.getLogger(__name__)


class CheckCvTermKeyValueValidator(ProfileValidationGroupValidator):
    def __init__(
        self,
        cv_helper: CvTermHelper,
        validators: dict[str, BaseProfileValidator],
        ftp_client_pool: dict[str, FTP],
        url_client: reachable.client.Client,
    ) -> None:
        super().__init__(cv_helper, validators, ftp_client_pool, url_client)

    def get_profile_validation_class(self) -> type[CheckCvTermKeyValue]:
        return CheckCvTermKeyValue

    def evaluate(
        self,
        value: Any,
        profile_validation: ProfileValidation,
        sub_path: None | list[int | str] = None,
    ) -> ValidationResult:
        if not isinstance(profile_validation, CheckCvTermKeyValue):
            return ValidationResult(
                sub_path=sub_path,
                name="",
                valid=False,
                message=f"Invalid validator {str(profile_validation)}",
                data=value,
            )

        validator: CheckCvTermKeyValue = profile_validation
        if not isinstance(value, list):
            return ValidationResult(
                sub_path=sub_path,
                name=validator.name,
                valid=False,
                message="Input must be list",
                data=value,
            )
        item: CvTermKeyValue = CvTermKeyValue.model_validate(value)
        if item.key.accession != validator.cv_term_key.accession:
            return ValidationResult(
                sub_path=sub_path,
                name=validator.name,
                valid=False,
                message=f"Not applicable to evaluate key {item.key.accession}",
                data=value,
            )
        return super().evaluate(
            value=item.values,
            profile_validation=profile_validation,
            sub_path=sub_path,
        )


class CheckCvChildTermKeyValuesValidator(BaseProfileValidator):
    def __init__(
        self,
        cv_helper: CvTermHelper,
        ftp_client_pool: dict[str, FTP],
        url_client: reachable.client.Client,
        validators: dict[str, BaseProfileValidator],
    ) -> None:
        super().__init__(cv_helper)
        self.ftp_client_pool = ftp_client_pool
        self.url_client = url_client
        self.validators = validators

    def get_profile_validation_class(self) -> type[CheckChildCvTermKeyValues]:
        return CheckChildCvTermKeyValues

    def evaluate(
        self,
        value: Any,
        profile_validation: ProfileValidation,
        sub_path: None | list[int | str] = None,
    ) -> ValidationResult:
        if not isinstance(profile_validation, CheckChildCvTermKeyValues):
            return ValidationResult(
                sub_path=sub_path,
                name="",
                valid=False,
                message=f"invalid validator {str(profile_validation)}",
                data=value,
            )

        validator: CheckChildCvTermKeyValues = profile_validation

        if (
            validator.conditional_field_name in value
            and validator.key_values_field_name in value
        ):
            # sub_path_copy = sub_path.copy()

            errors = []
            # for item in value:
            key = CvTerm.model_validate(value.get(validator.conditional_field_name))
            if str(key) == str(validator.conditional_cv_term):
                # sub_path_copy.append(validator.key_values_field_name)
                values = value.get(validator.key_values_field_name)
                key_value_validator = self.validators.get("check-cv-term-key-values")
                error = key_value_validator.validate(
                    values,
                    profile_validation.key_values_control,
                    sub_path=[validator.key_values_field_name],
                )
                if error:
                    error.message = f"{key.accession}, {key.name}: {error.message}"
                    errors.append(error)
            else:
                # sub_path_copy.append(validator.conditional_field_name)
                return ValidationResult(
                    sub_path=sub_path,
                    name=validator.name,
                    valid=True,
                    message=f"Key does not match: {key.name}, expected value {validator.conditional_cv_term.name}",
                    data=value,
                )
            if errors:
                return ValidationResult(
                    sub_path=sub_path,
                    name=validator.name,
                    valid=False,
                    message="Invalid controlled terms.",
                    data=value,
                    context=errors,
                )
            return ValidationResult(
                sub_path=sub_path,
                name=validator.name,
                valid=True,
                message="Valid controlled terms.",
                data=value,
            )


class CheckCvTermKeyValuesValidator(BaseProfileValidator):
    def __init__(
        self,
        cv_helper: CvTermHelper,
        ftp_client_pool: dict[str, FTP],
        url_client: reachable.client.Client,
        validators: dict[str, BaseProfileValidator],
    ) -> None:
        super().__init__(cv_helper)
        self.ftp_client_pool = ftp_client_pool
        self.url_client = url_client
        self.validators = validators

    def get_profile_validation_class(self) -> type[CheckCvTermKeyValues]:
        return CheckCvTermKeyValues

    def evaluate(
        self,
        value: Any,
        profile_validation: ProfileValidation,
        sub_path: None | list[int | str] = None,
    ) -> ValidationResult:
        if not isinstance(profile_validation, CheckCvTermKeyValues):
            return ValidationResult(
                sub_path=sub_path,
                name="",
                valid=False,
                message=f"invalid validator {str(profile_validation)}",
                data=value,
            )

        validator: CheckCvTermKeyValues = profile_validation
        if not isinstance(value, list):
            return ValidationResult(
                sub_path=sub_path,
                name=validator.name,
                valid=False,
                message="Input must be list",
                data=value,
            )
        items: list[CvTermKeyValue] = [CvTermKeyValue.model_validate(x) for x in value]

        errors: list[ValidationResult | ValidationError] = []
        for control_items, required in [
            (validator.required_items, True),
            (validator.optional_items, False),
        ]:
            if not control_items:
                continue

            control_map = {str(x.cv_term_key): x for x in control_items}
            validated_keys = set()
            invalid_keys = set()
            for item_idx, item in enumerate(items):
                item_key = str(item.key)
                if item_key not in control_map:
                    continue
                # sub_path_copy = sub_path.copy()
                # sub_path_copy.append(item_idx)
                key_validator = control_map[item_key]
                for idx, item_value in enumerate(item.values):
                    # sub_item_path = sub_path_copy.copy()
                    # sub_item_path.extend(["values"])
                    valid, invalid_results = self.validate_key_values(
                        key_validator, item_value, [item_idx, "values", idx]
                    )
                    if valid:
                        validated_keys.add(item_key)
                    else:
                        invalid_keys.add(item_key)
                        errors.append(
                            ValidationResult(
                                sub_path=[item_idx, "values", idx],
                                name=validator.name,
                                valid=False,
                                message=f"Item at index {idx} is not valid.",
                                data=value,
                                context=invalid_results,
                            )
                        )
            if required and len(validated_keys) != len(control_items):
                undefined_keys = [
                    f"[{control_map[x].cv_term_key.accession}, {control_map[x].cv_term_key.name}]"
                    for x in control_map
                    if x not in validated_keys and x not in invalid_keys
                ]
                invalid_values = [
                    control_map[x].cv_term_key.name
                    for x in control_map
                    if x in invalid_keys
                ]

                messages = []
                if undefined_keys:
                    messages.append(f"Undefined key(s): {', '.join(undefined_keys)}")
                if invalid_values:
                    messages.append(
                        f"Invalid value(s) in keys: {', '.join(invalid_values)}"
                    )
                errors.append(
                    ValidationResult(
                        sub_path=sub_path,
                        name=validator.name,
                        valid=False,
                        message=". ".join(messages),
                        data=value,
                    )
                )

        if errors:
            return ValidationResult(
                sub_path=sub_path,
                name=validator.name,
                valid=False,
                message="Invalid controlled terms. "
                + ". ".join([x.message for x in errors]),
                data=value,
                context=errors,
            )
        return ValidationResult(
            sub_path=sub_path,
            name=validator.name,
            valid=True,
            message="Valid controlled terms.",
            data=value,
        )

    def validate_key_values(
        self,
        key_validator: CheckCvTermKeyValue,
        value,
        sub_path: None | list[int | str],
    ):
        invalid_results: list[ValidationResult | ValidationError] = []
        for control in key_validator.controls:
            # sub_path_copy = sub_path.copy()
            validation_error = self.validators[control.name].validate(
                value, control.model_dump(by_alias=True), sub_path
            )

            if validation_error:
                validation_error.path = sub_path
                invalid_results.append(validation_error)

        valid = False
        valid_controls = len(key_validator.controls) - len(invalid_results)

        if key_validator.join_operator == "or":
            if valid_controls > 0:
                valid = True
        else:
            if valid_controls == len(key_validator.controls):
                valid = True

        return valid, invalid_results


class ProfileValidator:
    def __init__(self) -> None:
        ftp_client_pool: dict[str, FTP] = {}
        url_client: reachable.client.Client = reachable.client.Client()
        self.cv_helper = CvTermHelper()

        self.validators: dict[str, BaseProfileValidator] = {
            "allowed-cv-terms": AllowedCvTermValidator(self.cv_helper),
            "allow-any-cv-term": AllowAnyCvTermValidator(self.cv_helper),
            "allowed-children-cv-terms": AllowedChildrenCvTermsValidator(
                self.cv_helper
            ),
            "allowed-cv-list": AllowedCvListValidator(self.cv_helper),
            "accessible-uri": AccessibleURIValidator(
                self.cv_helper, ftp_client_pool=ftp_client_pool, url_client=url_client
            ),
            "accessible-compact-uri": AccessibleCompactURIValidator(
                self.cv_helper, ftp_client_pool=ftp_client_pool, url_client=url_client
            ),
        }
        self.validators["validation-group"] = ProfileValidationGroupValidator(
            self.cv_helper,
            validators=self.validators,
            ftp_client_pool=ftp_client_pool,
            url_client=url_client,
        )
        self.validators["check-cv-term-key-value"] = CheckCvTermKeyValueValidator(
            self.cv_helper,
            validators=self.validators,
            ftp_client_pool=ftp_client_pool,
            url_client=url_client,
        )
        self.validators["check-cv-term-key-values"] = CheckCvTermKeyValuesValidator(
            self.cv_helper,
            validators=self.validators,
            ftp_client_pool=ftp_client_pool,
            url_client=url_client,
        )

        self.validators["check-conditional-cv-term-key-values"] = (
            CheckCvChildTermKeyValuesValidator(
                self.cv_helper,
                validators=self.validators,
                ftp_client_pool=ftp_client_pool,
                url_client=url_client,
            )
        )

    def validate_instance(
        self, validator: jsonschema.Validator, profile_validation, instance, schema
    ):
        """
        Custom validation logic for the 'profile_validation' keyword.
        """
        if isinstance(profile_validation, dict) and "name" in profile_validation:
            model = ProfileValidation.model_validate(profile_validation)

            if model.list_join_operator in ("any", "minimum") and model.minimum is None:
                model.minimum = 1
            validator_name = model.name
            if validator_name in self.validators:
                validator_instance = self.validators[validator_name]
                if model.input_type in ("list-or-item", "item") and isinstance(
                    instance, list
                ):
                    validation_errors = []
                    valid_count = 0
                    valid = False
                    for idx, x in enumerate(instance):
                        # va = validator.evolve(schema=schema["items"])
                        # y = [x for x in va.iter_errors(x)]
                        validation_error = validator_instance.validate(
                            x, profile_validation, [idx]
                        )
                        if not validation_error:
                            valid_count += 1
                        else:
                            validation_errors.append(validation_error)
                        if model.list_join_operator == "all":
                            if validation_error:
                                yield validation_error
                        elif model.list_join_operator in ("any", "minimum"):
                            if model.minimum <= valid_count:
                                valid = True
                                break

                    if model.list_join_operator in ("any", "minimum") and not valid:
                        yield jsonschema.ValidationError(
                            message=f"Number of valid list items: {valid_count}, expected {model.minimum}.",
                            validator=model.name,
                            context=(),
                            path=(),
                            instance=instance,
                        )

                else:
                    validation_error = validator_instance.validate(
                        value=instance,
                        profile_validation=profile_validation,
                        sub_path=[],
                    )
                    if validation_error:
                        yield validation_error
            else:
                yield jsonschema.ValidationError(message="Not valid validation name")
        else:
            yield jsonschema.ValidationError(message="Not valid")

    @staticmethod
    def new_instance(
        schema_uri: None | str, profile_uri: None | str
    ) -> protocols.Validator:
        profile_validator = ProfileValidator()
        validator = validators.extend(
            jsonschema.Draft202012Validator,
            validators={"profileValidation": profile_validator.validate_instance},
        )
        if not schema_uri:
            schema_uri = SUPPORTED_SCHEMA_MAP.schemas[
                SUPPORTED_SCHEMA_MAP.default_schema_uri
            ]
        if schema_uri in SUPPORTED_SCHEMA_MAP.schemas:
            schema = SUPPORTED_SCHEMA_MAP.schemas.get(schema_uri)
            if not profile_uri:
                profile_uri = schema.default_profile_uri

            if schema.supported_profiles.get(profile_uri):
                _, schema_file = load_mhd_json_schema(profile_uri)
                return validator(schema_file)
        return None



import logging

import httpx

from mhd_model.shared.model import CvDefinition

logger = logging.getLogger(__name__)


def search_ontology_definition(ontology_name: str) -> None | CvDefinition:
    if not ontology_name:
        return None
    try:
        url = "https://www.ebi.ac.uk/ols4/api/v2/ontologies" + ontology_name.lower()
        response = httpx.get(url, timeout=2)
        response.raise_for_status()
        json_response = response.json()
        base_uri = json_response.get("baseUri", [])

        return CvDefinition(
            name=json_response.get("description", ""),
            uri=json_response.get("iri", ""),
            prefix=base_uri[0] if base_uri else "",
            label=json_response.get("preferredPrefix", "").upper(),
        )
    except Exception as e:
        logger.error(
            "Error while fetching ontology definition from OLS: '%s' - %s",
            ontology_name,
            e,
        )
        return None



import datetime

from pydantic import AnyUrl, EmailStr, Field, HttpUrl
from typing_extensions import Annotated

from mhd_model.model.v0_1.dataset.profiles.base.base import (
    BaseLabeledMhdModel,
    BasicCvTermModel,
    BasicCvTermValueModel,
    CvTermObjectId,
    CvTermValueObjectId,
    KeyValue,
    MhdObjectId,
    MhdObjectType,
)
from mhd_model.shared.fields import (
    DOI,
    ORCID,
    Authors,
    GrantId,
    PubMedId,
)
from mhd_model.shared.model import CvTermValue


class Person(BaseLabeledMhdModel):
    type_: Annotated[
        None | MhdObjectType,
        Field(
            frozen=True,
            alias="type",
            description="The value of this property MUST be 'person'",
        ),
    ] = "person"
    full_name: Annotated[
        None | str, Field(min_length=2, description="Full name of person")
    ] = None
    orcid: Annotated[
        None | ORCID,
        Field(
            title="ORCID",
            description="ORCID identifier of person",
            examples=["1234-0001-8473-1713", "1234-0001-8473-171X"],
        ),
    ] = None
    email_list: Annotated[
        None | list[EmailStr], Field(description="Email addresses of person")
    ] = None
    phone_list: Annotated[
        None | list[str],
        Field(
            description="Phone number of person (with international country code)",
            examples=[["+449340917271", "00449340917271"]],
        ),
    ] = None
    address_list: Annotated[
        None | list[str], Field(description="Addresses of person")
    ] = None

    def get_label(self):
        if self.orcid:
            return self.orcid
        if self.email_list and self.email_list[0]:
            return self.email_list[0]
        return self.full_name or self.id_


class Organization(BaseLabeledMhdModel):
    type_: Annotated[
        None | MhdObjectType,
        Field(
            frozen=True,
            description="The type property identifies type of the object",
            alias="type",
        ),
    ] = "organization"
    repository_identifier: Annotated[None | str, Field()] = None
    name: Annotated[str, Field(min_length=2)]
    department: Annotated[None | str, Field()] = None
    unit: Annotated[None | str, Field()] = None
    address: Annotated[None | str, Field()] = None


class Project(BaseLabeledMhdModel):
    type_: Annotated[
        None | MhdObjectType,
        Field(
            frozen=True,
            description="The type property identifies type of the object",
            alias="type",
        ),
    ] = "project"
    title: Annotated[None | str, Field(min_length=2)] = None
    description: Annotated[None | str, Field()] = None
    grant_identifier_list: Annotated[None | list[GrantId], Field()] = None
    doi: Annotated[None | DOI, Field()] = None

    def get_label(self):
        return self.title or self.id_


class Study(BaseLabeledMhdModel):
    type_: Annotated[
        None | MhdObjectType,
        Field(
            frozen=True,
            description="The type property identifies type of the object",
            alias="type",
        ),
    ] = "study"
    mhd_identifier: Annotated[None | str, Field()] = None
    repository_identifier: Annotated[None | str, Field(min_length=2)] = None
    additional_identifier_list: Annotated[None | list[CvTermValue], Field()] = None
    title: Annotated[None | str, Field()] = None
    description: Annotated[None | str, Field()] = None
    submission_date: None | datetime.datetime = None
    public_release_date: None | datetime.datetime = None
    license: Annotated[
        None | HttpUrl,
        Field(examples=[HttpUrl("https://creativecommons.org/publicdomain/zero/1.0/")]),
    ] = None
    grant_identifier_list: Annotated[None | list[GrantId], Field()] = None
    dataset_url_list: Annotated[None | list[AnyUrl], Field()] = None
    related_dataset_list: Annotated[None | list[KeyValue], Field()] = None
    protocol_refs: Annotated[None | list[MhdObjectId], Field()] = None

    def get_label(self):
        return self.mhd_identifier or self.title or self.id_


class Protocol(BaseLabeledMhdModel):
    type_: Annotated[
        None | MhdObjectType,
        Field(
            frozen=True,
            description="The type property identifies type of the object",
            alias="type",
        ),
    ] = "protocol"
    name: Annotated[None | str, Field()] = None
    protocol_type_ref: Annotated[None | CvTermObjectId, Field()] = None
    description: Annotated[None | str, Field()] = None
    parameter_definition_refs: Annotated[None | list[MhdObjectId], Field()] = None

    def get_label(self) -> str:
        return self.name or self.id_


class ParameterDefinition(BaseLabeledMhdModel):
    type_: Annotated[
        None | MhdObjectType,
        Field(
            frozen=True,
            description="The type property identifies type of the object",
            alias="type",
        ),
    ] = "parameter-definition"
    name: Annotated[None | str, Field()] = None
    parameter_type_ref: Annotated[None | CvTermObjectId, Field()] = None

    def get_label(self):
        return self.name or self.id_


class FactorDefinition(BaseLabeledMhdModel):
    type_: Annotated[
        None | MhdObjectType,
        Field(
            frozen=True,
            description="The type property identifies type of the object",
            alias="type",
        ),
    ] = "factor-definition"
    name: Annotated[None | str, Field()] = None
    factor_type_ref: Annotated[None | CvTermObjectId, Field()] = None

    def get_label(self):
        return self.name or self.id_


class CharacteristicDefinition(BaseLabeledMhdModel):
    type_: Annotated[
        None | MhdObjectType,
        Field(
            frozen=True,
            description="The type property identifies type of the object",
            alias="type",
        ),
    ] = "characteristic-definition"
    name: Annotated[None | str, Field()] = None
    characteristic_type_ref: Annotated[None | CvTermObjectId, Field()] = None

    def get_label(self):
        return self.name or self.id_


class Publication(BaseLabeledMhdModel):
    """
    A document that is the output of a publishing process. [IAO, IAO:0000311, publication]
    """

    type_: Annotated[
        None | MhdObjectType,
        Field(
            frozen=True,
            description="The type property identifies type of the object",
            alias="type",
        ),
    ] = "publication"
    title: Annotated[str, Field(min_length=10)]
    doi: Annotated[DOI, Field()]
    pubmed_id: Annotated[None | PubMedId, Field()] = None
    author_list: Annotated[None | Authors, Field()] = None

    def get_label(self):
        return self.doi or self.title or self.id_


class BasicAssay(BaseLabeledMhdModel):
    type_: Annotated[None | MhdObjectType, Field(..., frozen=True, alias="type")] = (
        "assay"
    )
    repository_identifier: Annotated[
        None | str,
        Field(
            description="An assay identifier that uniquely identifies the assay in repository."
        ),
    ] = None
    name: Annotated[
        None | str,
        Field(description="Name of the assay. It SHOULD be unique in a study."),
    ] = None
    metadata_file_ref: Annotated[
        None | MhdObjectId,
        Field(),
    ] = None

    technology_type_ref: Annotated[
        None | CvTermObjectId,
        Field(),
    ] = None
    assay_type_ref: Annotated[
        None | CvTermObjectId,
        Field(),
    ] = None
    measurement_type_ref: Annotated[
        None | CvTermObjectId,
        Field(),
    ] = None
    omics_type_ref: Annotated[
        None | CvTermObjectId,
        Field(),
    ] = None
    protocol_refs: Annotated[
        None | list[MhdObjectId],
        Field(
            description="The id properties of protocols used in assay. A protocol "
            "is a defined and standardized procedure followed to collect, prepare, "
            "or analyze biological samples.",
        ),
    ] = None

    def get_label(self):
        return self.name or self.id_


class Assay(BasicAssay):
    """[OBI, OBI:0000070, assay] A planned process that has the objective to produce information
    about a material entity by examining it.
    """

    type_: Annotated[None | MhdObjectType, Field(..., frozen=True, alias="type")] = (
        "assay"
    )
    sample_run_refs: Annotated[
        None | list[MhdObjectId],
        Field(),
    ] = None

    def get_label(self):
        return self.name or self.id_


class Subject(BaseLabeledMhdModel):
    type_: Annotated[
        None | MhdObjectType,
        Field(
            frozen=True,
            description="The type property identifies type of the object",
            alias="type",
        ),
    ] = "subject"
    name: Annotated[None | str, Field()] = None
    subject_type_ref: Annotated[None | CvTermObjectId, Field()] = None
    repository_identifier: Annotated[None | str, Field()] = None
    additional_identifier_list: Annotated[None | list[CvTermValue], Field()] = None

    def get_label(self):
        return self.name or self.id_


class Specimen(BaseLabeledMhdModel):
    type_: Annotated[
        None | MhdObjectType,
        Field(
            frozen=True,
            description="The type property identifies type of the object",
            alias="type",
        ),
    ] = "specimen"
    name: Annotated[None | str, Field()] = None
    repository_identifier: Annotated[None | str, Field()] = None
    additional_identifier_list: Annotated[None | list[CvTermValue], Field()] = None

    def get_label(self):
        return self.name or self.id_


class Sample(BaseLabeledMhdModel):
    type_: Annotated[
        None | MhdObjectType,
        Field(
            frozen=True,
            description="The type property identifies type of the object",
            alias="type",
        ),
    ] = "sample"
    name: Annotated[None | str, Field()] = None
    repository_identifier: Annotated[None | str, Field()] = None
    additional_identifier_list: Annotated[None | list[CvTermValue], Field()] = None

    def get_label(self):
        return self.name or self.id_


class SampleRun(BaseLabeledMhdModel):
    type_: Annotated[
        None | MhdObjectType,
        Field(
            frozen=True,
            description="The type property identifies type of the object",
            alias="type",
        ),
    ] = "sample-run"
    name: Annotated[None | str, Field()] = None
    sample_ref: Annotated[
        None | MhdObjectId,
        Field(),
    ] = None
    sample_run_configuration_refs: Annotated[
        None | list[MhdObjectId],
        Field(),
    ] = None
    raw_data_file_refs: Annotated[
        None | list[MhdObjectId],
        Field(),
    ] = None
    derived_data_file_refs: Annotated[
        None | list[MhdObjectId],
        Field(),
    ] = None
    result_file_refs: Annotated[
        None | list[MhdObjectId],
        Field(),
    ] = None
    supplementary_file_refs: Annotated[
        None | list[MhdObjectId],
        Field(),
    ] = None

    def get_label(self):
        return self.name or self.id_


class SampleRunConfiguration(BaseLabeledMhdModel):
    type_: Annotated[
        None | MhdObjectType,
        Field(
            frozen=True,
            description="The type property identifies type of the object",
            alias="type",
        ),
    ] = "sample-run-configuration"
    protocol_ref: Annotated[
        None | MhdObjectId,
        Field(),
    ] = None
    parameter_value_refs: Annotated[
        None | list[MhdObjectId | CvTermObjectId | CvTermValueObjectId],
        Field(),
    ] = None


class Metabolite(BaseLabeledMhdModel):
    """
    Any intermediate or product resulting from metabolism.
    The term 'metabolite' subsumes the classes commonly known as primary and secondary metabolites. [CHEBI, CHEBI:25212, metabolite]
    """

    type_: Annotated[
        None | MhdObjectType,
        Field(
            frozen=True,
            description="The type property identifies type of the object",
            alias="type",
        ),
    ] = "metabolite"
    name: Annotated[None | str, Field()] = None

    def get_label(self):
        return self.name or self.id_


class BaseFile(BaseLabeledMhdModel):
    name: Annotated[
        None | str,
        Field(
            description="Name of the file. File MUST be a file (not folder or link)."
            "It MAY be relative path "
            "(e.g., FILES/study.txt) or a file in a compressed file "
            "(e.g., FILES/study.zip#data/metadata.tsv)."
        ),
    ] = None
    size: Annotated[
        None | int,
        Field(
            description="The size of the file in bytes, "
            "representing the total amount of data contained in the file."
        ),
    ] = None
    hash_sha256: Annotated[
        None | str,
        Field(
            description="The SHA-256 cryptographic hash of the file content, "
            "used to verify file integrity and ensure that the file has not been altered."
        ),
    ] = None
    format_ref: Annotated[
        None | CvTermObjectId,
        Field(
            description="The structure or encoding used to store the contents of the file, "
            "typically indicated by its extension (e.g., .txt, .csv, .mzML, .raw, etc.)."
        ),
    ] = None
    compression_format_refs: Annotated[
        None | list[CvTermObjectId],
        Field(
            description="The structure or encoding used to compress the contents of the file, "
            "typically indicated by its extension (e.g., .zip, .tar, .gz, etc.)."
            " List item order shows order of compressions. e.g. [tar format, gzip format] for tar.gz"
        ),
    ] = None
    extension: Annotated[
        None | str,
        Field(
            description="The extension of file. It MUST contain all extensions "
            "(e.g., .raw, .mzML, .d.zip, .raw.zip, etc.)."
        ),
    ] = None

    def get_label(self):
        return self.name or self.id_


class ReferencedDataFile(BaseFile):
    pass
    # metadata_file_refs: Annotated[
    #     None | list[MhdObjectId],
    #     Field(
    #         description="The id properties of metadata file that references or describes the file."
    #     ),
    # ] = None


class RawDataFile(ReferencedDataFile):
    """[MS, MS:1003083, raw data file]
    Data file that contains original data as generated by an instrument,
    although not necessarily in the original data format
    (i.e. an original raw file converted to a different format is still a raw data file)
    """

    type_: Annotated[
        None | MhdObjectType,
        Field(
            frozen=True,
            description="The type property identifies type of the object",
            alias="type",
        ),
    ] = "raw-data-file"


class DerivedDataFile(ReferencedDataFile):
    """[MS, MS:1003084, processed data file]
    File that contains data that has been substantially processed or
    transformed from what was originally acquired by an instrument.
    """

    type_: Annotated[
        None | MhdObjectType,
        Field(
            frozen=True,
            description="The type property identifies type of the object",
            alias="type",
        ),
    ] = "derived-data-file"


class MetadataFile(BaseFile):
    type_: Annotated[
        None | MhdObjectType,
        Field(
            frozen=True,
            description="The type property identifies type of the object",
            alias="type",
        ),
    ] = "metadata-file"


class ResultFile(BaseFile):
    type_: Annotated[
        None | MhdObjectType,
        Field(
            frozen=True,
            description="The type property identifies type of the object",
            alias="type",
        ),
    ] = "result-file"
    # metadata_file_refs: Annotated[None | list[MhdObjectId], Field()] = None


class SupplementaryFile(BaseFile):
    type_: Annotated[
        None | MhdObjectType,
        Field(
            frozen=True,
            description="The type property identifies type of the object",
            alias="type",
        ),
    ] = "supplementary-file"


class CvTermObject(BasicCvTermModel):
    type_: Annotated[
        None | MhdObjectType,
        Field(
            alias="type",
            description="The type property identifies type of the CV Term object",
        ),
    ] = "cv-term"


class CvTermValueObject(BasicCvTermValueModel):
    type_: Annotated[
        None | MhdObjectType,
        Field(
            alias="type",
            description="The type property identifies type of the CV Term Value object",
        ),
    ] = "cv-term-value"



from typing import Annotated, Any, Literal

from pydantic import Field, ValidationError, field_validator

from mhd_model.shared.model import CvTerm
from mhd_model.shared.validation.registry import (
    VALIDATORS,
    ProfileValidation,
    ValidatorBaseModel,
)


class CvTermPlaceholder(ValidatorBaseModel):
    source: str = ""
    accession: str = ""

    def __hash__(self) -> int:
        return f"[{self.source}, {self.accession}]"


class ProfileCvTermValidation(ProfileValidation):
    allowed_missing_cv_terms: None | list[CvTerm] = None
    allowed_placeholder_values: None | list[CvTermPlaceholder] = None
    allowed_other_sources: None | list[str] = None

    def __str__(self) -> str:
        result = [
            (
                "Allowed Missing CV Terms: "
                + ", ".join([str(x) for x in self.allowed_missing_cv_terms])
                if self.allowed_missing_cv_terms
                else None
            ),
            (
                "Allowed Placeholder Values: "
                + ", ".join([str(x) for x in self.allowed_placeholder_values])
                if self.allowed_placeholder_values
                else None
            ),
            (
                "Allowed Other Sources: "
                + ", ".join([x for x in self.allowed_other_sources])
                if self.allowed_other_sources
                else None
            ),
        ]
        filtered_result = [x for x in result if x]
        return "Exceptions:\n" + "\n".join(filtered_result) if filtered_result else ""


class AllowedCvList(ProfileCvTermValidation):
    name: str = "allowed-cv-list"
    source_names: list[str]

    def __str__(self) -> str:
        sources = (
            "Ontology Sources:" + ", ".join(self.source_names)
            if self.source_names
            else ""
        )
        parent_str = super().__str__()

        if parent_str:
            return sources + f"\n{parent_str}"

        return sources


class ParentCvTerm(ValidatorBaseModel):
    cv_term: CvTerm
    allow_only_leaf: bool = False
    allow_parent: None | bool = False
    excluded_cv_terms: None | list[CvTerm] = None
    index_cv_terms: None | bool = True

    def __str__(self) -> str:
        parent = str(self.cv_term)
        excludes = (
            "Excluded CV Terms: " + ", ".join([str(x) for x in self.excluded_cv_terms])
            if self.excluded_cv_terms
            else ""
        )
        allow_parent = "Allow parent CV Term: " + ("Yes" if self.allow_parent else "No")
        allow_only_leaf_terms = "Allow only leaf CV Terms: " + (
            "Yes" if self.allow_only_leaf else "No"
        )

        result = [
            x for x in [parent, allow_parent, allow_only_leaf_terms, excludes] if x
        ]
        return "\n".join(result)


class AllowedChildrenCvTerms(ProfileCvTermValidation):
    name: str = "allowed-children-cv-terms"
    parent_cv_terms: list[ParentCvTerm]

    def __str__(self) -> str:
        sources = (
            "Allowed Parent CV Terms: "
            + ", ".join([str(x) for x in self.parent_cv_terms])
            if self.parent_cv_terms
            else ""
        )
        parent_str = super().__str__()

        if parent_str:
            return sources + f"\n{parent_str}"

        return sources


class AllowedCvTerms(ProfileCvTermValidation):
    name: Annotated[str, Field()] = "allowed-cv-terms"
    cv_terms: None | list[CvTerm]

    def __str__(self) -> str:
        sources = (
            "Allowed CV Terms: " + ", ".join([str(x) for x in self.cv_terms])
            if self.cv_terms
            else ""
        )
        parent_str = super().__str__()

        if parent_str:
            return sources + f"\n{parent_str}"

        return sources


class AccessibleURI(ProfileValidation):
    name: str = "accessible-uri"
    max_retries: int = 1
    retry_period_in_seconds: int = 0


class AccessibleCompactURI(ProfileValidation):
    name: str = "accessible-compact-uri"
    default_prefix: None | str = None
    follow_redirects: bool = False


class AllowAnyCvTerm(ProfileCvTermValidation):
    name: str = "allow-any-cv-term"

    def __str__(self) -> str:
        sources = "Allow any valid CV Term"
        parent_str = super().__str__()

        if parent_str:
            return sources + f"\n{parent_str}"

        return sources


class ProfileValidationGroup(ProfileValidation):
    name: str = "validation-group"
    controls: list[ProfileValidation]
    join_operator: Literal["and", "or"] = "and"

    @field_validator("controls", mode="before")
    @classmethod
    def controls_validator(cls, value: Any) -> None:
        if not value:
            return []
        return_value = []
        if isinstance(value, list):
            for item in value:
                if isinstance(item, ProfileValidation):
                    return_value.append(item)
                elif isinstance(item, dict) and "name" in item:
                    validation_name = item["name"]
                    validator = VALIDATORS.get(validation_name)
                    if validator:
                        return_value.append(validator.model_validate(item))
                    else:
                        raise ValidationError(
                            f"invalid validator name {validation_name}"
                        )

        else:
            raise ValidationError(f"invalid value {value}")
        return return_value



import sys

import click

from mhd_model.commands.create.announcement import create_announcement_file_task
from mhd_model.commands.create.neo4j_input import create_neo4j_input_file_task
from mhd_model.commands.create.sdrf import create_sdrf_file_task


@click.group(name="create", context_settings={"help_option_names": ["-h", "--help"]})
def create_group():
    """utilities to create and convert files (announcement file, etc.)."""
    pass


create_group.add_command(create_announcement_file_task)
create_group.add_command(create_sdrf_file_task)
create_group.add_command(create_neo4j_input_file_task)


if __name__ == "__main__":
    if len(sys.argv) == 1:
        create_group(["--help"])
    else:
        create_group()



import json
from pathlib import Path
from typing import Any

import click

from mhd_model.model.definitions import (
    MHD_MODEL_V0_1_DEFAULT_SCHEMA_NAME,
    MHD_MODEL_V0_1_LEGACY_PROFILE_NAME,
    MHD_MODEL_V0_1_MS_PROFILE_NAME,
)
from mhd_model.model.v0_1.dataset.profiles.legacy.profile import MhDatasetLegacyProfile
from mhd_model.model.v0_1.dataset.profiles.ms.profile import MhDatasetMsProfile
from mhd_model.shared.model import ProfileEnabledDataset


@click.command(name="neo4j-input", no_args_is_help=True)
@click.option(
    "--output-dir",
    default="outputs",
    show_default=True,
    help="Output directory for neo4j input file",
)
@click.option(
    "--output-filename",
    default=None,
    show_default=True,
    help="neo4j input filename (e.g., MHD000001.neo4j_input.json, ST000001.neo4j_input.json)",
)
@click.argument("mhd_study_id")
@click.argument("mhd_model_file_path")
def create_neo4j_input_file_task(
    mhd_study_id: str,
    mhd_model_file_path: str,
    output_dir: str,
    output_filename: str,
):
    """Create neo4j input file from MHD data model file.

    Args:

    mhd_study_id (str): MHD study identifier

    mhd_model_file_path (str): MHD data model path


    output_dir (str): Output directory of neo4j input file

    output_filename (str): Name of neo4j input file. Default is <repository identifier>.neo4j_input.json
    """
    file = Path(mhd_model_file_path)

    Path(output_dir).mkdir(parents=True, exist_ok=True)

    txt = file.read_text()
    json_data = json.loads(txt)

    profile: ProfileEnabledDataset = ProfileEnabledDataset.model_validate(json_data)
    mhd_dataset = None
    if profile.schema_name == MHD_MODEL_V0_1_DEFAULT_SCHEMA_NAME:
        if profile.profile_uri == MHD_MODEL_V0_1_LEGACY_PROFILE_NAME:
            mhd_dataset = MhDatasetLegacyProfile.model_validate(json_data)
        elif profile.profile_uri == MHD_MODEL_V0_1_MS_PROFILE_NAME:
            mhd_dataset = MhDatasetMsProfile.model_validate(json_data)
        else:
            click.echo(f"{profile.profile_uri} is not supported.")
            exit(1)
    else:
        click.echo(f"{profile.schema_name} is not schema.")
        exit(1)

    nodes_map = {x.id_: x for x in mhd_dataset.graph.nodes}
    relationships_map = {x.id_: x for x in mhd_dataset.graph.relationships}

    nodes: list[dict[str, Any]] = []
    relationships: list[dict[str, Any]] = []
    for node in nodes_map.values():
        properties = {}
        for key, value in node.model_dump(exclude_none=True).items():
            refs = []
            if key == "id_":
                continue
            if key.endswith("_ref"):
                ref = getattr(node, key, None)
                if hasattr(nodes_map[ref], "name"):
                    val = getattr(nodes_map[ref], "name")
                    properties[key.replace("_ref", "")] = val
                refs.append(ref)
            elif key.endswith("_refs"):
                refs = getattr(node, key, [])
                vals = []
                for ref in refs:
                    if hasattr(nodes_map[ref], "name"):
                        val = getattr(nodes_map[ref], "name")
                        if val and val not in vals:
                            vals.append(val)
                properties[key.replace("_refs", "") + "_list"] = vals
            if refs:
                link_name = "EMBEDDED_" + key.replace("_ref", "").replace(
                    "_refs", ""
                ).upper().replace("-", "_")
                for ref in refs:
                    relationships.append(
                        {
                            "start": node.id_,
                            "end": ref,
                            "type": link_name,
                            "properties": {},
                        }
                    )

            if not refs:
                if isinstance(value, list):
                    properties[key] = [str(x) for x in value]
                elif isinstance(value, dict):
                    properties[key] = json.dumps({k: str(v) for k, v in value.items()})
                elif (
                    isinstance(value, str)
                    or isinstance(value, int)
                    or isinstance(value, float)
                    or isinstance(value, bool)
                ):
                    properties[key] = value
                else:
                    properties[key] = str(value)

        properties["label"] = node.label

        nodes.append(
            {
                "id": node.id_,
                "labels": [node.type_],
                "properties": properties,
            }
        )

    for rel in relationships_map.values():
        # if nodes_map[rel.source_ref].type_ == "study" and nodes_map[
        #     rel.target_ref
        # ].type_ in {
        #     "raw-data-file",
        #     "derived-data-file",
        #     "supplementary-file",
        #     # "metabolite",
        # }:
        #     continue
        # if nodes_map[rel.target_ref].type_ == "study" and nodes_map[
        #     rel.source_ref
        # ].type_ in {
        #     "raw-data-file",
        #     "derived-data-file",
        #     "supplementary-file",
        #     # "metabolite",
        # }:
        #     continue
        relationships.append(
            {
                "start": rel.source_ref,
                "end": rel.target_ref,
                "type": rel.relationship_name.upper().replace("-", "_"),
                "properties": rel.model_dump(exclude_none=True),
            }
        )
    output_file = f"{output_dir}/{mhd_study_id}.neo4j_input.json"
    if output_filename:
        output_file = f"{output_dir}/{output_filename}"

    with Path(output_file).open("w") as f:
        json.dump({"nodes": nodes, "relationships": relationships}, f, indent=2)

    click.echo(f"{output_file} is created.")



from __future__ import annotations

import abc
import logging
import pathlib
from ftplib import FTP
from typing import Any
from urllib.parse import urlparse

import bioregistry
import httpx
import jsonschema
import reachable
from pydantic import BaseModel, ValidationError

from mhd_model.shared.model import CvTerm, CvTermKeyValue, CvTermValue
from mhd_model.shared.validation.cv_term_helper import (
    CvTermHelper,
)
from mhd_model.shared.validation.definitions import (
    AccessibleCompactURI,
    AccessibleURI,
    AllowAnyCvTerm,
    AllowedChildrenCvTerms,
    AllowedCvList,
    AllowedCvTerms,
    CvTermPlaceholder,
    ProfileCvTermValidation,
    ProfileValidationGroup,
)
from mhd_model.shared.validation.registry import ProfileValidation

logger = logging.getLogger(__name__)


class ValidationResult(BaseModel):
    name: str
    root: bool = False
    message: None | str
    valid: bool = False
    data: None | Any = None
    sub_path: None | list[int | str] = None
    error: None | Any = None
    context: None | Any = None
    parent: None | Any = None
    validator_value: None | Any = None


class BaseProfileValidator(abc.ABC):
    def __init__(self, cv_helper: CvTermHelper) -> None:
        self.cv_helper = cv_helper

    def is_accessible_url(self, url: str) -> bool:
        try:
            result = httpx.head(url, timeout=5)
            if result.status_code == 404:
                return False
            result.raise_for_status()
            return True
        except Exception as ex:
            logger.debug("Unaccessible URL: %s", str(ex))
            return False

    @abc.abstractmethod
    def get_profile_validation_class(self) -> type[ProfileValidation]:
        pass

    def validate(
        self,
        value: Any,
        profile_validation: dict[str, Any],
        sub_path: None | list[int | str] = None,
    ) -> tuple[bool, list[ValidationResult]]:
        if isinstance(profile_validation, dict):
            validation = self.get_profile_validation_class().model_validate(
                profile_validation
            )
        else:
            validation = profile_validation

        if not value:
            if validation.allow_null_value:
                return None

        if isinstance(validation, ProfileCvTermValidation):
            if isinstance(value, CvTermKeyValue) or (
                isinstance(value, dict) and value.get("key")
            ):
                data = value
                if isinstance(value, dict):
                    data = CvTermKeyValue.model_validate(value)
                if validation.allowed_placeholder_values:
                    if data.values and data.values[0].accession in {
                        x.accession for x in validation.allowed_placeholder_values
                    }:
                        return None
                if validation.allowed_missing_cv_terms:
                    if data.values and data.values[0] in {
                        x.accession for x in validation.allowed_missing_cv_terms
                    }:
                        return None

            if isinstance(value, CvTerm) or (
                isinstance(value, dict) and "source" in value and "accession" in value
            ):
                data = value
                if isinstance(value, dict):
                    data = CvTermPlaceholder.model_validate(value)

                if validation.allowed_placeholder_values:
                    if str(data) in {
                        str(x) for x in validation.allowed_placeholder_values
                    }:
                        return None
                if validation.allowed_missing_cv_terms:
                    if data.accession in {
                        x.accession for x in validation.allowed_missing_cv_terms
                    }:
                        return None
                if validation.allowed_other_sources:
                    if data.accession in {x for x in validation.allowed_other_sources}:
                        if not data.source:
                            source = data.accession.split(":")[0]
                            data.source = source
                        url = self.cv_helper.get_uri(cv_term=data)
                        acessible = self.is_accessible_url(url=url)
                        if acessible:
                            return None
                        error = jsonschema.ValidationError(
                            message=f"{data.accession} is not accessible",
                            validator=validation.name,
                            context=[],
                            path=(),
                            instance=data,
                        )

        validation_result = self.evaluate(value, validation, [])
        valid = validation_result is None or validation_result.valid

        if validation.negate:
            valid = not valid

        if not valid:
            # sub_path = sub_path if sub_path else []
            # new_subpath = validation_result.sub_path if validation_result.sub_path else []
            # if new_subpath:
            #     sub_path.extend(new_subpath)
            error = jsonschema.ValidationError(
                # message=self.create_message(validation_result),
                message=validation_result.message,
                validator=validation.name,
                context=[],
                path=(),
                instance=value,
            )
            self.update_error_context(error, validation_result)
            return error
        return None

    def update_error_context(
        self, error: jsonschema.ValidationError, result: ValidationResult
    ) -> None:
        context = []
        if result.context:
            for item in result.context:
                if isinstance(item, jsonschema.ValidationError):
                    context.append(item)
                elif isinstance(item, ValidationResult) and not item.valid:
                    cnxt_error = jsonschema.ValidationError(
                        # message=self.create_message(validation_result),
                        message=item.message,
                        validator=item.name,
                        context=item.context if item.context else [],
                        path=item.sub_path if item.sub_path else (),
                        instance=item.data,
                    )
                    self.update_error_context(cnxt_error, item)
                    context.append(cnxt_error)
        error.context = context

    @abc.abstractmethod
    def evaluate(
        self,
        value: Any,
        profile_validation: ProfileValidation,
        sub_path: None | list[int | str] = None,
    ) -> ValidationResult:
        pass

    def create_message(
        self, error: jsonschema.ValidationError | ValidationResult
    ) -> str:
        if not error:
            return "validation error"
        if isinstance(error, jsonschema.ValidationError):
            name = error.validator
            message = error.message
            jsonpath_str = self.json_path(error.absolute_path)
        else:
            name = error.name
            message = error.message
            jsonpath_str = ""
        if error.context and len(error.context) == 1:
            message = f"{message}"
        else:
            message = f"{name}: {message}"
        if error.context:
            sub_messages = []
            for x in error.context:
                sub_messages.append(self.create_message(x))
            if not jsonpath_str:
                return message + f" [{', '.join(sub_messages)}]"
            return message + f" [{jsonpath_str}: {', '.join(sub_messages)}]"
        return message

    def json_path(self, field_path: list[str | int]) -> str:
        return ".".join([x if isinstance(x, str) else f"[{x}]" for x in field_path])


class AllowedCvTermValidator(BaseProfileValidator):
    def __init__(self, cv_helper: CvTermHelper) -> None:
        super().__init__(cv_helper)

    def get_profile_validation_class(self) -> type[AllowedCvTerms]:
        return AllowedCvTerms

    def evaluate(
        self,
        value: Any,
        profile_validation: ProfileValidation,
        sub_path: None | list[int | str] = None,
    ) -> ValidationResult:
        try:
            val: CvTerm = CvTerm.model_validate(value)
        except ValidationError as ex:
            ex.errors()

        validator: AllowedCvTerms = profile_validation

        for term in validator.cv_terms:
            if (
                val.name == term.name
                and val.accession == term.accession
                and val.source == term.source
            ):
                message = None
                return ValidationResult(
                    sub_path=sub_path,
                    name=validator.name,
                    valid=True,
                    message=f"[{val.source}, {val.accession}, {val.name}] is an allowed CV term.",
                    data=value,
                )
        terms = ", ".join(
            [f"[{x.source}, {x.accession}, {x.name}]" for x in validator.cv_terms]
        )
        message = f"[{val.source}, {val.accession}, {val.name}] does not match any allowed CV term. Allowed CV terms: {terms}"

        return ValidationResult(
            sub_path=sub_path,
            name=validator.name,
            valid=False,
            message=message,
            data=value,
        )


class AllowAnyCvTermValidator(BaseProfileValidator):
    def __init__(self, cv_helper: CvTermHelper) -> None:
        super().__init__(cv_helper)

    def get_profile_validation_class(self) -> type[AllowAnyCvTerm]:
        return AllowAnyCvTerm

    def evaluate(
        self,
        value: Any,
        profile_validation: ProfileValidation,
        sub_path: None | list[int | str] = None,
    ) -> ValidationResult:
        validator: AllowAnyCvTerm = profile_validation
        data = CvTerm.model_validate(value)
        if isinstance(data, CvTerm):
            if ":" not in data.accession:
                return ValidationResult(
                    sub_path=sub_path,
                    name=validator.name,
                    message=f"{data.accession} is not valid CURIE value.",
                    data=value,
                )

            # prefix, identifier = tuple(data.accession.split(":"))

            result = self.cv_helper.get_uri(data)
            if not result:
                return ValidationResult(
                    sub_path=sub_path,
                    name=validator.name,
                    message=f"[{data.name}, {data.accession}, {data.name}] is not valid CV term",
                    data=value,
                )
        else:
            return ValidationResult(
                sub_path=sub_path,
                name=validator.name,
                message=f"{value} is not an valid CV term",
                data=value,
            )
        return ValidationResult(
            sub_path=sub_path,
            valid=True,
            name=validator.name,
            message=f"[{data.name}, {data.accession}, {data.name}] is valid CV term",
            data=value,
        )


class ProfileValidationGroupValidator(BaseProfileValidator):
    def __init__(
        self,
        cv_helper: CvTermHelper,
        validators: dict[str, BaseProfileValidator],
        ftp_client_pool: dict[str, FTP],
        url_client: reachable.client.Client,
    ) -> None:
        super().__init__(cv_helper)
        self.ftp_client_pool = ftp_client_pool
        self.url_client = url_client
        self.validators = validators

    def get_profile_validation_class(self) -> type[ProfileValidationGroup]:
        return ProfileValidationGroup

    def evaluate(
        self,
        value: Any,
        profile_validation: ProfileValidation,
        sub_path: None | list[int | str] = None,
    ) -> ValidationResult:
        validation_tests: list[ValidationResult | list[ValidationResult]] = []
        profile_validation_list: ProfileValidationGroup = profile_validation
        invalid_results = []

        if isinstance(value, list):
            list_invalid_results = []
            for control in profile_validation_list.controls:
                for idx, item in enumerate(value):
                    sub_path_copy = sub_path.copy()
                    sub_path_copy.append(idx)
                    validation_error = self.validators[control.name].validate(
                        item, control, [idx]
                    )
                    if validation_error:
                        list_invalid_results.extend(validation_error)
                if list_invalid_results:
                    validation_tests.append((control.name, list_invalid_results))
        else:
            for control in profile_validation_list.controls:
                sub_path_copy = sub_path.copy()
                validation_error = self.validators[control.name].validate(
                    value, control, []
                )

                if validation_error:
                    invalid_results.append((control.name, validation_error))

        valid = False
        valid_controls = len(profile_validation_list.controls) - len(invalid_results)

        if profile_validation_list.join_operator == "or":
            if valid_controls > 0:
                valid = True
        else:
            if valid_controls == len(profile_validation_list.controls):
                valid = True

        validation_results = [x[1] for x in invalid_results]

        if valid:
            return ValidationResult(
                sub_path=sub_path,
                name=profile_validation_list.name,
                valid=True,
                message=f"{value} is valid.",
                data=value,
                context=validation_results,
            )
        return ValidationResult(
            sub_path=sub_path,
            name=profile_validation_list.name,
            valid=False,
            message=f"{value} is not valid.",
            data=value,
            context=validation_results,
        )


class AllowedChildrenCvTermsValidator(BaseProfileValidator):
    def __init__(self, cv_helper: CvTermHelper) -> None:
        super().__init__(cv_helper)

    def get_profile_validation_class(self) -> type[AllowedChildrenCvTerms]:
        return AllowedChildrenCvTerms

    def evaluate(
        self,
        value: Any,
        profile_validation: ProfileValidation,
        sub_path: None | list[int | str] = None,
    ) -> ValidationResult:
        validator: AllowedChildrenCvTerms = profile_validation
        if not isinstance(profile_validation, AllowedChildrenCvTerms):
            ValidationResult(
                sub_path=sub_path,
                name=validator.name,
                valid=False,
                message="Invalid validator",
                data=value,
            )

        parents = ", ".join(
            [
                f"[{x.cv_term.source}, {x.cv_term.accession}, {x.cv_term.name}]"
                for x in validator.parent_cv_terms
            ]
        )
        cv_term = value
        if isinstance(cv_term, dict):
            cv_term: CvTerm = CvTerm.model_validate(value)
        if not isinstance(cv_term, CvTerm):
            return ValidationResult(
                sub_path=sub_path,
                name=validator.name,
                valid=False,
                message="input is not Cv term",
                data=value,
            )
        for parent in validator.parent_cv_terms:
            if parent.index_cv_terms:
                terms = self.cv_helper.get_children_of_cv_term(parent)
                if cv_term.accession in terms:
                    return ValidationResult(
                        sub_path=sub_path,
                        name=validator.name,
                        valid=True,
                        message=f"[{cv_term.source}, {cv_term.accession}, {cv_term.name}] is a child of allowed CV terms. ",
                        data=value,
                    )

            else:
                search_result, message = self.cv_helper.check_cv_term(
                    cv_term=cv_term, parent_cv_term=parent
                )
                if search_result:
                    return ValidationResult(
                        sub_path=sub_path,
                        name=validator.name,
                        valid=True,
                        message=f"[{cv_term.source}, {cv_term.accession}, {cv_term.name}] is a child of allowed CV terms. ",
                        data=value,
                    )
                return ValidationResult(
                    sub_path=sub_path,
                    name=validator.name,
                    valid=search_result,
                    message=message,
                    data=value,
                )

        return ValidationResult(
            sub_path=sub_path,
            name=validator.name,
            valid=False,
            message=f"[{cv_term.source}, {cv_term.accession}, {cv_term.name}] is not child of allowed CV terms. "
            f"Terms should be child of any {parents} CV terms",
            data=value,
        )


class AllowedCvListValidator(BaseProfileValidator):
    def __init__(self, cv_helper: CvTermHelper) -> None:
        super().__init__(cv_helper)

    def get_profile_validation_class(self) -> type[AllowedCvList]:
        return AllowedCvList

    def evaluate(
        self,
        value: Any,
        profile_validation: ProfileValidation,
        sub_path: None | list[int | str] = None,
    ) -> ValidationResult:
        if not isinstance(profile_validation, AllowedCvList):
            return ValidationResult(
                sub_path=sub_path,
                name="invalid validation ",
                valid=False,
                message="Invalid validator",
                data=value,
            )
        data = value
        if isinstance(value, dict):
            data = CvTerm.model_validate(value)
        validator: AllowedCvList = profile_validation

        if data.source in validator.source_names:
            check_valid, message = self.cv_helper.check_cv_term(data)
            if check_valid:
                return ValidationResult(
                    sub_path=sub_path,
                    name=validator.name,
                    valid=True,
                    message=f"[{data.source}, {data.accession}, {data.name}] is an allowed CV source.",
                    data=value,
                )
            return ValidationResult(
                sub_path=sub_path,
                name=validator.name,
                valid=check_valid,
                message=message,
                data=value,
            )
        else:
            labels = ", ".join(validator.source_names)
            message = (
                f"[{data.source}, {data.accession}, {data.name}] is not in any allowed CV source. "
                f"Define a CV term from the selected CV sources: {labels}"
            )
            return ValidationResult(
                sub_path=sub_path,
                name=validator.name,
                valid=False,
                message=message,
                data=value,
            )


class AccessibleURIValidator(BaseProfileValidator):
    def __init__(
        self,
        cv_helper: CvTermHelper,
        ftp_client_pool: dict[str, FTP],
        url_client: reachable.client.Client,
    ) -> None:
        super().__init__(cv_helper)
        self.ftp_client_pool = ftp_client_pool
        self.url_client = url_client
        self.cache: dict[None | str, ValidationResult] = {}

    def get_profile_validation_class(self) -> type[AccessibleURI]:
        return AccessibleURI

    def return_result(self, key: str, result: ValidationResult) -> ValidationResult:
        if key not in self.cache:
            self.cache[key] = result
        return result

    def evaluate(
        self,
        value: Any,
        profile_validation: ProfileValidation,
        sub_path: None | list[int | str] = None,
    ) -> ValidationResult:
        validator: AccessibleURI = profile_validation
        try:
            data: CvTermValue = CvTermValue.model_validate(value)

            if not data.value:
                logger.debug("URL is not defined")
                return self.return_result(
                    data.value,
                    ValidationResult(
                        sub_path=sub_path,
                        name=validator.name,
                        valid=False,
                        message="URL is not defined",
                        data=value,
                    ),
                )
            if not isinstance(data.value, str):
                logger.debug("Value type is not string: %s", data.value)
                sub_path.append("value")
                return self.return_result(
                    data.value,
                    ValidationResult(
                        sub_path=sub_path,
                        name=validator.name,
                        valid=False,
                        message=f"Value type is not string: {data.value}",
                        data=value,
                    ),
                )
            url: str = data.value

            uri = urlparse(url=url)
            updated_url = url.split("#")[0]
            if updated_url in self.cache:
                return self.cache[updated_url]

            if uri.scheme in ("http", "https"):
                sub_path.append("value")
                logger.debug("Check URL: %s", updated_url)
                result = httpx.head(updated_url, timeout=5)
                if result.status_code == 404:
                    logger.debug("URL does not exist: %s", url)
                    return self.return_result(
                        updated_url,
                        ValidationResult(
                            sub_path=sub_path,
                            name=validator.name,
                            valid=False,
                            message=f"URL does not exist: {value}",
                            data=value,
                        ),
                    )
                result.raise_for_status()
                logger.debug("URL exist: %s", url)
                return self.return_result(
                    updated_url,
                    ValidationResult(
                        sub_path=sub_path,
                        name=validator.name,
                        valid=True,
                        message=f"URL exists: {value}",
                        data=value,
                    ),
                )
            elif uri.scheme in ("ftp", "ftp"):
                sub_path.append("value")
                if uri.hostname not in self.ftp_client_pool:
                    ftp = FTP(user="", host=uri.hostname, timeout=5)
                    ftp.login()
                    if uri.hostname not in self.ftp_client_pool:
                        self.ftp_client_pool[uri.hostname] = ftp
                ftp = self.ftp_client_pool[uri.hostname]
                try:
                    full_path = pathlib.Path(uri.path)
                    ftp.size(str(full_path))

                except Exception as ex:
                    logger.debug(
                        "Target file is not accessible: %s - %s",
                        url,
                        ex.__qualname__,
                    )
                    return self.return_result(
                        updated_url,
                        ValidationResult(
                            sub_path=sub_path,
                            name=validator.name,
                            valid=False,
                            message=f"Target file is not accessible: {url}. {ex.__qualname__}",
                            data=value,
                        ),
                    )

                logger.debug("URL exist: %s", url)
                return self.return_result(
                    updated_url,
                    ValidationResult(
                        sub_path=sub_path,
                        name=validator.name,
                        valid=True,
                        message=f"URL exists: {url}",
                        data=value,
                    ),
                )
            sub_path.append("accession")
            logger.error("Unsupported protocol %s", uri.scheme)
            return self.return_result(
                updated_url,
                ValidationResult(
                    sub_path=sub_path,
                    name=validator.name,
                    valid=False,
                    message=f"Unsupported protocol {url}",
                    data=value,
                ),
            )
        except Exception as ex:
            logger.error("Unaccessible URI: %s", str(ex))
            return self.return_result(
                updated_url,
                ValidationResult(
                    sub_path=sub_path,
                    name=validator.name,
                    valid=False,
                    message=f"Unaccessible URI: {value}",
                    data=value,
                ),
            )


class AccessibleCompactURIValidator(BaseProfileValidator):
    def __init__(
        self,
        cv_helper: CvTermHelper,
        ftp_client_pool: dict[str, FTP],
        url_client: reachable.client.Client,
    ) -> None:
        super().__init__(cv_helper)
        self.ftp_client_pool = ftp_client_pool
        self.url_client = url_client

    def get_profile_validation_class(self) -> type[AccessibleCompactURI]:
        return AccessibleCompactURI

    def evaluate(
        self,
        value: Any,
        profile_validation: ProfileValidation,
        sub_path: None | list[int | str] = None,
    ) -> ValidationResult:
        if not isinstance(profile_validation, AccessibleCompactURI):
            return ValidationResult(
                sub_path=sub_path,
                name="",
                valid=False,
                message=f"invalid validator {str(profile_validation)}",
                data=value,
            )

        if value is None:
            if profile_validation.allow_null_value:
                return ValidationResult(
                    sub_path=sub_path,
                    name=profile_validation.name,
                    valid=True,
                    message="Allow empty value",
                    data=value,
                )
            return ValidationResult(
                sub_path=sub_path,
                name=profile_validation.name,
                valid=False,
                message="Empty file is not allowed.",
                data=value,
            )
        validator: AccessibleCompactURI = profile_validation

        if not validator.default_prefix:
            return ValidationResult(
                sub_path=sub_path,
                name=validator.name,
                valid=False,
                message=f"Invalid URI  prefix: {validator.default_prefix}",
                data=value,
            )
        identifier = value.replace(f"{validator.default_prefix}:", "")
        default_uri = bioregistry.get_default_iri(validator.default_prefix, identifier)

        if not default_uri:
            return ValidationResult(
                sub_path=sub_path,
                name=validator.name,
                valid=False,
                message=f"Invalid URI: {default_uri}",
                data=value,
            )

        headers = {
            "Accept": "*/*",
            "Accept-Encoding": "gzip, deflate, br",
            "Connection": "keep-alive",
            "Cache-Control": "no-cache",
            "User-Agent": "PostmanRuntime/7.43.3",
            "Authorization": "",
        }
        try:
            url: str = default_uri
            result = httpx.get(
                url,
                headers=headers,
                follow_redirects=validator.follow_redirects,
                timeout=60,
            )
            if not validator.follow_redirects:
                if result.status_code in (301, 302):
                    logger.debug("URL is redirected: %s", default_uri)
                    return ValidationResult(
                        sub_path=sub_path,
                        name=validator.name,
                        valid=True,
                        message=f"URL is redirected: {default_uri}",
                        data=value,
                    )
            if result.status_code == 404:
                logger.warning("URL is not found: %s", default_uri)
                return ValidationResult(
                    sub_path=sub_path,
                    name=validator.name,
                    valid=False,
                    message=f"URL is not found: {default_uri}",
                    data=value,
                )
            result.raise_for_status()
            return ValidationResult(
                sub_path=sub_path,
                name=validator.name,
                valid=True,
                message=f"URL exists: {default_uri}",
                data=value,
            )

        except Exception as ex:
            logger.debug("Unaccessible URI: %s, %s", default_uri)
            logger.exception(ex)
            return ValidationResult(
                sub_path=sub_path,
                name=validator.name,
                valid=False,
                message=f"Unaccessible URI: {default_uri}",
                data=value,
            )



from pydantic import Field
from typing_extensions import Annotated

from mhd_model.model.v0_1.dataset.profiles.base.base import (
    MhdObjectType,
)
from mhd_model.model.v0_1.dataset.profiles.base.profile import (
    MhDatasetBaseProfile,
)


class MhDatasetMsProfile(MhDatasetBaseProfile):
    type_: Annotated[MhdObjectType, Field(frozen=True, alias="type")] = MhdObjectType(
        "ms-dataset"
    )



import logging
from typing import Any

import jsonschema
import jsonschema.protocols
from jsonschema import protocols, validators

from mhd_model.model.definitions import (
    MHD_MODEL_V0_1_LEGACY_PROFILE_NAME,
    MHD_MODEL_V0_1_MS_PROFILE_NAME,
    SUPPORTED_SCHEMA_MAP,
)
from mhd_model.model.v0_1.dataset.profiles.legacy.graph_validation import (
    MHD_LEGACY_PROFILE_V0_1,
)
from mhd_model.model.v0_1.dataset.profiles.ms.graph_validation import (
    MHD_MS_PROFILE_V0_1,
)
from mhd_model.model.v0_1.dataset.validation.base import MhdModelValidator
from mhd_model.schema_utils import load_mhd_json_schema
from mhd_model.shared.exceptions import MhdValidationError
from mhd_model.shared.model import ProfileEnabledDataset
from mhd_model.utils import json_path, load_json

logger = logging.getLogger(__name__)


def validate_mhd_file(file_path: str):
    json_data = load_json(file_path)

    mhd_validator = MhdFileValidator()
    errors = mhd_validator.validate(json_data)

    messages = set()
    validation_errors = []
    for x in errors:
        if x.message in messages:
            continue
        messages.add(x.message)
        validation_errors.append((json_path(x.absolute_path), x))
    validation_errors.sort(key=lambda x: x[0])
    return validation_errors


MHD_PROFILE_VALIDATIONS_V0_1 = {
    MHD_MODEL_V0_1_MS_PROFILE_NAME: MHD_MS_PROFILE_V0_1,
    MHD_MODEL_V0_1_LEGACY_PROFILE_NAME: MHD_LEGACY_PROFILE_V0_1,
}


def new_validator(
    schema_uri: None | str, profile_uri: None | str
) -> protocols.Validator:
    if not schema_uri:
        schema_uri = SUPPORTED_SCHEMA_MAP.schemas[
            SUPPORTED_SCHEMA_MAP.default_schema_uri
        ]
    if schema_uri in SUPPORTED_SCHEMA_MAP.schemas:
        schema = SUPPORTED_SCHEMA_MAP.schemas.get(schema_uri)
        if not profile_uri:
            profile_uri = schema.default_profile_uri

        if (
            schema.supported_profiles.get(profile_uri)
            and profile_uri in MHD_PROFILE_VALIDATIONS_V0_1
        ):
            _, schema_file = load_mhd_json_schema(profile_uri)
            node_validation = MHD_PROFILE_VALIDATIONS_V0_1[profile_uri]

            mhd_model_validator = MhdModelValidator(node_validation)

            validator = validators.extend(
                jsonschema.Draft202012Validator,
                validators={
                    "mhdGraphValidation": mhd_model_validator.validate_graph,
                    "anyOf": mhd_model_validator.anyOf,
                },
            )
            logger.info("Loaded schema: %s, profile: %s.", schema_uri, profile_uri)
            return validator(schema_file)
    return None


def get_profile(schema_uri: None | str, profile_uri: None | str) -> protocols.Validator:
    if not schema_uri:
        schema_uri = SUPPORTED_SCHEMA_MAP.schemas[
            SUPPORTED_SCHEMA_MAP.default_schema_uri
        ]
    if schema_uri in SUPPORTED_SCHEMA_MAP.schemas:
        schema = SUPPORTED_SCHEMA_MAP.schemas.get(schema_uri)
        if not profile_uri:
            profile_uri = schema.default_profile_uri

        if (
            schema.supported_profiles.get(profile_uri)
            and profile_uri in MHD_PROFILE_VALIDATIONS_V0_1
        ):
            return MHD_PROFILE_VALIDATIONS_V0_1[profile_uri]

    return None


class MhdFileValidator:
    def validate(self, json_file: dict[str, Any]) -> list[jsonschema.ValidationError]:
        profile: ProfileEnabledDataset = ProfileEnabledDataset.model_validate(json_file)
        validator: jsonschema.protocols.Validator = new_validator(
            profile.schema_name, profile.profile_uri
        )
        if not validator:
            logger.error(
                "No validator found for schema %s with profile URI %s",
                profile.schema_name,
                profile.profile_uri,
            )
            raise MhdValidationError(
                f"No validator found for schema {profile.schema_name} with profile URI {profile.profile_uri}"
            )
        validations = validator.iter_errors(json_file)
        all_errors = [x for x in validations]
        return all_errors



import abc
from pathlib import Path

from mhd_model.shared.model import Revision


class BaseMhdConvertor(abc.ABC):
    @abc.abstractmethod
    def convert(
        self,
        reposipory_name: str,
        repository_identifier: str,
        mhd_identifier: None | str,
        mhd_output_folder_path: Path,
        repository_revision: None | Revision = None,
        **kwargs,
    ): ...


class BaseMhdConvertorFactory(abc.ABC):
    @abc.abstractmethod
    def get_convertor(
        self,
        target_mhd_model_schema_uri: str,
        target_mhd_model_profile_uri: str,
    ) -> BaseMhdConvertor: ...



from typing import Any

from pydantic import Field, field_validator
from pydantic.alias_generators import to_pascal
from typing_extensions import Annotated

from mhd_model.model.v0_1.dataset.profiles.base import base, graph_nodes, relationships
from mhd_model.model.v0_1.dataset.profiles.base.base import (
    BaseMhdModel,
    BaseMhdRelationship,
    CvTermObjectId,
    CvTermValueObjectId,
    IdentifiableMhdModel,
    MhdConfigModel,
    MhdObjectId,
    MhdObjectType,
)
from mhd_model.model.v0_1.dataset.profiles.base.relationships import Relationship
from mhd_model.shared.model import CvEnabledDataset


class MhdGraph(MhdConfigModel):
    start_item_refs: Annotated[
        list[MhdObjectId | CvTermObjectId | CvTermValueObjectId], Field()
    ] = []
    nodes: Annotated[
        list[
            graph_nodes.CvTermValueObject
            | graph_nodes.CvTermObject
            | graph_nodes.Person
            | graph_nodes.Project
            | graph_nodes.Study
            | graph_nodes.Protocol
            | graph_nodes.Publication
            | graph_nodes.BasicAssay
            | graph_nodes.Assay
            | graph_nodes.Specimen
            | graph_nodes.Subject
            | graph_nodes.Sample
            | graph_nodes.SampleRun
            | graph_nodes.SampleRunConfiguration
            | graph_nodes.Metabolite
            | graph_nodes.MetadataFile
            | graph_nodes.ResultFile
            | graph_nodes.RawDataFile
            | graph_nodes.DerivedDataFile
            | graph_nodes.SupplementaryFile
            | graph_nodes.BaseLabeledMhdModel
        ],
        Field(),
    ] = []
    relationships: Annotated[list[Relationship], Field()] = []

    @field_validator("nodes", mode="before")
    @classmethod
    def node_validator(cls, v) -> list[BaseMhdModel]:
        if isinstance(v, list):
            items = []
            for item in v:
                if isinstance(item, BaseMhdModel):
                    items.append(item)
                elif isinstance(item, dict):
                    val = cls.create_model(item)
                    if not val:
                        raise ValueError("invalid type in nodes")
                    items.append(val)
                else:
                    raise ValueError("invalid type in nodes")
            return items

        raise ValueError("invalid type")

    @field_validator("relationships", mode="before")
    @classmethod
    def relationship_validator(cls, v) -> list[BaseMhdRelationship]:
        if isinstance(v, list):
            items = []
            for item in v:
                if isinstance(item, BaseMhdRelationship):
                    items.append(item)
                elif isinstance(item, dict):
                    class_name = to_pascal(item["type"].replace("-", "_"))
                    if hasattr(relationships, class_name):
                        class_object = getattr(relationships, class_name)
                        items.append(class_object.model_validate(item))
                else:
                    raise ValueError("invalid type in nodes")
            return items

        raise ValueError("invalid type")

    @staticmethod
    def create_model(item: dict[str, Any]):
        class_name = to_pascal(item["type"].replace("-", "_"))
        if item["id"].startswith("cv--"):
            return graph_nodes.CvTermObject.model_validate(item)
        elif item["id"].startswith("cv-value--"):
            return graph_nodes.CvTermValueObject.model_validate(item)
        elif hasattr(graph_nodes, class_name):
            class_object = getattr(graph_nodes, class_name)
            return class_object.model_validate(item)
        return None

    @staticmethod
    def get_node_class(item: dict[str, Any]):
        if not item or not item.get("type") or not item.get("id"):
            return None

    @staticmethod
    def get_mhd_class_by_type(node_type: str) -> None | IdentifiableMhdModel:
        class_name = to_pascal(node_type.replace("-", "_"))
        class_object = None
        if hasattr(graph_nodes, class_name):
            class_object = getattr(graph_nodes, class_name)
        if not class_object:
            class_object = getattr(base, class_name)
        return class_object


class GraphEnabledBaseDataset(CvEnabledDataset): ...


class MhDatasetBaseProfile(GraphEnabledBaseDataset):
    id_: Annotated[
        None | str,
        Field(
            alias="id",
            description="Unique identifier of graph node",
        ),
    ] = None
    type_: Annotated[MhdObjectType, Field(frozen=True, alias="type")] = MhdObjectType(
        "base-dataset"
    )
    name: Annotated[None | str, Field()] = None
    description: Annotated[None | str, Field()] = None
    graph: Annotated[MhdGraph, Field(json_schema_extra={"mhdGraphValidation": {}})] = (
        MhdGraph()
    )



from typing import Literal

from pydantic import BaseModel, ConfigDict
from pydantic.alias_generators import to_camel, to_pascal


class ValidatorBaseModel(BaseModel):
    """Base model class to convert python attributes to camel case"""

    model_config = ConfigDict(
        populate_by_name=True,
        JSON_schema_serialization_defaults_required=True,
        field_title_generator=lambda field_name, field_info: to_pascal(
            field_name.replace("_", " ").strip()
        ),
        alias_generator=to_camel,
    )


class ProfileValidation(ValidatorBaseModel):
    name: str
    negate: bool = False
    allow_null_value: bool = False
    list_join_operator: Literal["any", "all", "minimum"] = "all"
    minimum: None | int = None
    input_type: Literal["list-or-item", "item", "list"] = "list-or-item"


def register_validator_class(
    validator_name: str, model_class: type[ProfileValidation]
) -> None:
    VALIDATORS[validator_name] = model_class


def unregister_validator_class(validator_name: str) -> None:
    validator = VALIDATORS.get(validator_name)
    if validator:
        del VALIDATORS[validator_name]


VALIDATORS: dict[str, type[ProfileValidation]] = {}



import datetime

from pydantic import AnyUrl, EmailStr, Field, HttpUrl
from typing_extensions import Annotated

from mhd_model.model.v0_1.announcement.profiles.base import fields
from mhd_model.shared.fields import Authors
from mhd_model.shared.model import (
    CvEnabledDataset,
    CvTerm,
    MhdConfigModel,
)


class AnnouncementBaseModel(MhdConfigModel):
    """Base model for announcement-related models."""

    pass


class AnnouncementBaseFile(AnnouncementBaseModel):
    name: Annotated[str, Field(min_length=1)]
    url_list: Annotated[list[AnyUrl], Field(min_length=1)]
    compression_formats: Annotated[None | list[fields.CompressionFormat], Field()] = (
        None
    )
    extension: Annotated[None | str, Field(min_length=2)] = None


class AnnouncementMetadataFile(AnnouncementBaseFile):
    format: Annotated[None | fields.MetadataFileFormat, Field()] = None


class AnnouncementRawDataFile(AnnouncementBaseFile):
    format: Annotated[None | fields.RawDataFileFormat, Field()] = None


class AnnouncementResultFile(AnnouncementBaseFile):
    format: Annotated[None | fields.ResultFileFormat, Field()] = None


class AnnouncementDerivedDataFile(AnnouncementBaseFile):
    format: Annotated[None | fields.DerivedFileFormat, Field()] = None


class AnnouncementSupplementaryFile(AnnouncementBaseFile):
    format: Annotated[None | fields.SupplementaryFileFormat, Field()] = None


class AnnouncementContact(AnnouncementBaseModel):
    """A contact associated with the dataset.
    This can be a submitter or a principal investigator.
    """

    full_name: Annotated[None | str, Field(min_length=5)] = None
    emails: Annotated[None | list[EmailStr], Field(min_length=1)] = None
    orcid: Annotated[None | fields.ORCID, Field(title="ORCID")] = None
    affiliations: Annotated[None | list[str], Field(min_length=1)] = None


class AnnouncementPublication(AnnouncementBaseModel):
    """A publication associated with the dataset."""

    title: Annotated[str, Field(min_length=10)]
    doi: Annotated[fields.DOI, Field()]
    pubmed_id: Annotated[None | fields.PubMedId, Field()] = None
    author_list: Annotated[None | Authors, Field()] = None


class AnnouncementReportedMetabolite(AnnouncementBaseModel):
    name: Annotated[str, Field(min_length=1)]
    database_identifiers: Annotated[
        None | list[fields.MetaboliteDatabaseId], Field()
    ] = None


class AnnouncementProtocol(AnnouncementBaseModel):
    """A protocol is a defined and standardized procedure followed
    to collect, prepare, or analyze biological samples.
    """

    name: Annotated[str, Field()]
    protocol_type: Annotated[fields.ProtocolType, Field()]
    description: Annotated[None | str, Field()] = None
    protocol_parameters: Annotated[
        None | list[fields.ExtendedCvTermKeyValue], Field()
    ] = None
    relates_assay_names: Annotated[None | list[str], Field()] = None


class AnnouncementBaseProfile(CvEnabledDataset, AnnouncementBaseModel):
    """Base profile for announcement files."""

    mhd_identifier: Annotated[None | str, Field()] = None
    repository_identifier: Annotated[str, Field()]
    mhd_metadata_file_url: Annotated[AnyUrl, Field()]
    dataset_url_list: Annotated[list[AnyUrl], Field(min_length=1)]

    license: Annotated[None | HttpUrl, Field()] = None
    title: Annotated[str, Field(min_length=25)]
    description: Annotated[None | str, Field(min_length=60)]
    submission_date: Annotated[None | datetime.datetime, Field()]
    public_release_date: Annotated[None | datetime.datetime, Field()]

    submitters: Annotated[None | list[AnnouncementContact], Field(min_length=1)]
    principal_investigators: Annotated[None | list[AnnouncementContact], Field()] = None

    # Metabolomics, Lipidomics, Proteomics, ...
    omics_type: Annotated[None | list[fields.OmicsType], Field(min_length=1)] = None
    # NMR, MS, ...
    technology_type: Annotated[
        None | list[fields.TechnologyType], Field(min_length=1)
    ] = None
    # Targeted metabolite profiling, Untargeted metabolite profiling, ...
    measurement_type: Annotated[None | list[fields.MeasurementType], Field()] = None
    # LC-MS, GC-MS, ...
    assay_type: Annotated[None | list[fields.AssayType], Field(min_length=1)] = None

    submitter_keywords: Annotated[None | list[fields.CvTermOrStr], Field()] = None
    descriptors: Annotated[None | list[CvTerm], Field()] = None

    publications: Annotated[
        None | fields.MissingPublicationReason | list[AnnouncementPublication],
        Field(),
    ] = None

    study_factors: Annotated[None | list[fields.ExtendedCvTermKeyValue], Field()] = None

    characteristic_values: Annotated[
        None | list[fields.ExtendedCvTermKeyValue], Field()
    ] = None

    protocols: Annotated[
        None | list[AnnouncementProtocol],
        Field(description="The protocols used in the study."),
    ] = None

    reported_metabolites: Annotated[
        None | list[AnnouncementReportedMetabolite], Field()
    ] = None

    repository_metadata_file_list: Annotated[
        None | list[AnnouncementMetadataFile], Field(min_length=1)
    ] = None
    raw_data_file_list: Annotated[
        None | list[AnnouncementRawDataFile], Field(min_length=1)
    ] = None
    derived_data_file_list: Annotated[
        None | list[AnnouncementDerivedDataFile], Field(min_length=1)
    ] = None
    supplementary_file_list: Annotated[
        None | list[AnnouncementSupplementaryFile],
        Field(min_length=1),
    ] = None
    result_file_list: Annotated[
        None | list[AnnouncementResultFile], Field(min_length=1)
    ] = None



from typing import Annotated, Any

from pydantic import Field, model_validator

from mhd_model.shared.model import MhdConfigModel


class EmbeddedRefValidation(MhdConfigModel):
    node_type: Annotated[None | str, Field()] = None
    node_property_name: Annotated[None | str, Field()] = None
    required: Annotated[None | bool, Field()] = None
    target_ref_types: Annotated[list[str], Field()]
    min: Annotated[None | int, Field()] = None
    max: Annotated[None | int, Field()] = None

    def __str__(self) -> str:
        targets = (
            "Target node type: <code>**"
            + ", ".join(self.target_ref_types)
            + "**</code>"
            if self.target_ref_types
            else ""
        )
        min_length = None
        max_length = None
        if self.min:
            min_length = f"Minimum length: **{self.min}**"

        if self.max:
            max_length = f"Maximum length: **{self.max}**"
        return "\n".join([x for x in [targets, min_length, max_length] if x])


class RelationshipValidation(MhdConfigModel):
    description: Annotated[None | str, Field()] = None
    source: Annotated[None | str, Field()]
    source_property_name: Annotated[None | str, Field()] = None
    relationship_name: Annotated[str, Field()]
    reverse_relationship_name: Annotated[None | str, Field()]
    target: Annotated[None | str, Field()]
    min: Annotated[int, Field()]
    max: Annotated[None | int, Field()] = None
    min_for_each_source: Annotated[None | int, Field()] = None
    max_for_each_source: Annotated[None | int, Field()] = None

    @model_validator(mode="wrap")
    @classmethod
    def validate_model(cls, v: Any, handler: callable) -> "RelationshipValidation":
        item: RelationshipValidation = handler(v)
        if item.source is not None:
            item.source = item.source.lower().replace(" ", "-")
        if item.target is not None:
            item.target = item.target.lower().replace(" ", "-")
        item.relationship_name = item.relationship_name.lower().replace(" ", "-")
        return item


class RequiredRelationshipValidation(RelationshipValidation):
    description: Annotated[None | str, Field()] = None
    source: Annotated[None | str, Field()]
    relationship_name: Annotated[str, Field()]
    reverse_relationship_name: Annotated[None | str, Field()]
    target: Annotated[None | str, Field()]
    min_for_each_source: Annotated[int, Field()] = 1
    expression: Annotated[str, Field()]
    expected_value: Annotated[str, Field()]



from mhd_model.shared.model import CvDefinition

CONTROLLED_CV_DEFINITIONS = {
    "MTBLS": CvDefinition(
        label="MTBLS",
        name="MetaboLights Controlled Vocabulary",
        uri="https://raw.githubusercontent.com/EBI-Metabolights/MtblsWS-Py/refs/heads/main/resources/Metabolights.owl",
        prefix="http://www.ebi.ac.uk/metabolights/ontology/MTBLS_",
    ),
    "BTO": CvDefinition(
        label="BTO",
        name="The BRENDA Tissue Ontology (BTO)",
        uri="http://purl.obolibrary.org/obo/bto.owl",
        prefix="http://purl.obolibrary.org/obo/BTO_",
    ),
    "CHEBI": CvDefinition(
        label="CHEBI",
        name="Chemical Entities of Biological Interest",
        uri="http://purl.obolibrary.org/obo/chebi.owl",
        prefix="http://purl.obolibrary.org/obo/CHEBI_",
    ),
    "CHEMINF": CvDefinition(
        label="CHEMINF",
        name="chemical information ontology (cheminf) - information entities about chemical entities",
        uri="http://semanticchemistry.github.io/semanticchemistry/ontology/cheminf.owl",
        prefix="http://semanticscience.org/resource/CHEMINF_",
    ),
    "CHMO": CvDefinition(
        label="CHMO",
        name="Chemical Methods Ontology",
        uri="http://purl.obolibrary.org/obo/chmo.owl",
        prefix="http://purl.obolibrary.org/obo/CHMO_",
    ),
    "DOID": CvDefinition(
        label="DOID",
        name="Human Disease Ontology",
        uri="http://purl.obolibrary.org/obo/doid.owl",
        prefix="http://purl.obolibrary.org/obo/DOID_",
    ),
    "EDAM": CvDefinition(
        label="EDAM",
        name="The ontology of data analysis and management",
        uri="http://edamontology.org",
        prefix="http://edamontology.org/format_",
        alternative_prefixes=[
            "http://edamontology.org/operation_",
            "http://edamontology.org/topic_, http://edamontology.org/data_",
        ],
    ),
    "EFO": CvDefinition(
        label="EFO",
        name="Experimental Factor Ontology",
        uri="http://www.ebi.ac.uk/efo/efo.owl",
        prefix="http://www.ebi.ac.uk/efo/EFO_",
    ),
    "ENVO": CvDefinition(
        label="ENVO",
        name="The Environment Ontology",
        uri="http://purl.obolibrary.org/obo/envo.owl",
        prefix="http://purl.obolibrary.org/obo/ENVO_",
    ),
    "GO": CvDefinition(
        label="GO",
        name="Gene Ontology",
        uri="http://purl.obolibrary.org/obo/go/extensions/go-plus.owl",
        prefix="http://purl.obolibrary.org/obo/GO_",
    ),
    "MI": CvDefinition(
        label="MI",
        name="Molecular Interactions Controlled Vocabulary",
        uri=" http://purl.obolibrary.org/obo/mi.owl",
        prefix="http://purl.obolibrary.org/obo/MI_",
    ),
    "MS": CvDefinition(
        label="MS",
        name="MS",
        uri="http://purl.obolibrary.org/obo/ms.owl",
        prefix="http://purl.obolibrary.org/obo/MS_",
    ),
    "MSIO": CvDefinition(
        label="MSIO",
        name="Metabolomics Standards Initiative Ontology (MSIO)",
        uri="http://purl.obolibrary.org/obo/msio.owl",
        prefix="http://purl.obolibrary.org/obo/MSIO_",
    ),
    "NCBITAXON": CvDefinition(
        label="NCBITAXON",
        name="National Center for Biotechnology Information (NCBI) Organismal Classification",
        uri="http://purl.obolibrary.org/obo/ncbitaxon.owl",
        prefix="http://purl.obolibrary.org/obo/NCBITaxon_",
        alternative_labels=["NCBITaxon"],
        alternative_prefixes=["http://purl.bioontology.org/ontology/NCBITAXON/"],
    ),
    "NCIT": CvDefinition(
        label="NCIT",
        name="NCI Thesaurus OBO Edition",
        uri="http://purl.obolibrary.org/obo/ncit.owl",
        prefix="http://purl.obolibrary.org/obo/NCIT_",
    ),
    "OBI": CvDefinition(
        label="OBI",
        name="Ontology for Biomedical Investigations",
        uri="http://purl.obolibrary.org/obo/obi.owl",
        prefix="http://purl.obolibrary.org/obo/OBI_",
    ),
    "SWO": CvDefinition(
        label="SWO",
        name="Software Ontology",
        uri="http://www.ebi.ac.uk/swo/swo.owl",
        prefix="http://www.ebi.ac.uk/swo/SWO_",
    ),
}


OTHER_CONTROLLED_CV_DEFINITIONS = {
    "DOI": CvDefinition(
        label="DOI",
        name="Digital Object Identifier",
        uri="https://www.doi.org",
        prefix="https://www.doi.org/",
        alternative_labels=["doi"],
        alternative_prefixes=["http://www.doi.org/"],
    ),
    "ORCID": CvDefinition(
        label="ORCID",
        name="Digital Object Identifier",
        uri="https://orcid.org",
        prefix="https://orcid.org/",
        alternative_labels=["orcid"],
        alternative_prefixes=["http://orcid.org/"],
    ),
    "PMID": CvDefinition(
        label="PMID",
        name="PubMed",
        uri="https://pubmed.ncbi.nlm.nih.gov",
        prefix="https://pubmed.ncbi.nlm.nih.gov/",
        alternative_labels=["pmid"],
        alternative_prefixes=["http://pubmed.ncbi.nlm.nih.gov/"],
    ),
    "WIKIDATA": CvDefinition(
        label="WIKIDATA",
        name="Wikidata",
        uri="https://www.wikidata.org/wiki",
        prefix="https://www.wikidata.org/wiki/",
        alternative_labels=["wikidata"],
        alternative_prefixes=["http://www.wikidata.org/wiki/"],
    ),
    "REFMET": CvDefinition(
        label="REFMET",
        name="Reference list of Metabolite names",
        uri="https://www.metabolomicsworkbench.org/databases/refmet/refmet_details.php",
        prefix="https://www.metabolomicsworkbench.org/databases/refmet/refmet_details.php?REFMET_ID=",
        alternative_labels=["refmet", "REFMET"],
    ),
    "ILX": CvDefinition(
        label="ILX",
        name="InterLex Identifier",
        uri="http://www.interlex.org",
        prefix="http://uri.interlex.org/user/ilx_",
        alternative_labels=["ilx"],
        alternative_prefixes=["http://uri.interlex.org/user/ilx_"],
    ),
    "WORMS": CvDefinition(
        label="WORMS",
        name="World Register of Marine Species",
        uri="https://www.marinespecies.org",
        prefix="https://www.marinespecies.org/aphia.php?p=taxdetails&id=",
        alternative_labels=["WORMS", "worms", "WoRMS"],
        alternative_prefixes=[
            "http://www.marinespecies.org/aphia.php?p=taxdetails&id="
        ],
    ),
}



from __future__ import annotations

import json
from pathlib import Path

import click

from mhd_model.model.v0_1.dataset.validation.validator import validate_mhd_file
from mhd_model.shared.model import ProfileEnabledDataset


@click.command(name="mhd", no_args_is_help=True)
@click.option(
    "--output-path",
    default=None,
    help="Validation output file path",
)
@click.argument("mhd_study_id")
@click.argument("mhd_model_file_path")
def validate_mhd_file_task(
    mhd_study_id: str,
    mhd_model_file_path: str,
    output_path: None | str,
):
    """Validate MHD model file.

    Args:

    mhd_study_id (str): MHD study id

    mhd_model_file_path (str): MHD model file path

    output_path (None | str): If it is not None, validation results are saved in output file path.
    """
    file = Path(mhd_model_file_path)
    try:
        txt = file.read_text()
        announcement_file_json = json.loads(txt)
        profile: ProfileEnabledDataset = ProfileEnabledDataset.model_validate(
            announcement_file_json
        )
        click.echo(
            f"{mhd_study_id}: {mhd_model_file_path} MHD file validation started."
        )
        click.echo(f"Used schema: {profile.schema_name}")
        click.echo(f"Validation profile: {profile.profile_uri}")
        all_errors = validate_mhd_file(mhd_model_file_path)

    except Exception as ex:
        import traceback

        traceback.print_exc()
        all_errors = [("exception", str(ex))]

    errors_list = []
    for key, error in all_errors:
        if hasattr(error, "message"):
            errors_list.append({"key": key, "error": error.message})
        else:
            errors_list.append({"key": key, "error": error})

    if output_path:
        output_file = Path(output_path)
        output_file.parent.mkdir(parents=True, exist_ok=True)
        with output_file.open("w") as f:
            result = {
                "success": len(errors_list) == 0,
                "errors": [str(x) for x in errors_list],
            }
            json.dump(result, f, indent=4)
    if not errors_list:
        click.echo(
            f"{mhd_study_id}: File '{mhd_model_file_path}' is validated successfully."
        )
        exit(0)

    click.echo("")
    click.echo(
        f"{mhd_study_id}: {mhd_model_file_path} has ({len(errors_list)}) validation errors."
    )
    click.echo("-" * 100)
    for idx, item in enumerate(errors_list, start=1):
        click.echo(f"{idx}\t{item.get('key')}\t{item.get('error')}")
    click.echo("-" * 100)
    exit(1)



import json
from pathlib import Path

import click

from mhd_model.convertors.announcement.v0_1.legacy.mhd2announce import (
    create_announcement_file,
)


@click.command(name="announcement", no_args_is_help=True)
@click.option(
    "--output-dir",
    default="outputs",
    show_default=True,
    help="Output directory for MHD file",
)
@click.option(
    "--output-filename",
    default=None,
    show_default=True,
    help="MHD announcement filename (e.g., MHD000001.announcement.json, ST000001.announcement.json)",
)
@click.argument("mhd_study_id")
@click.argument("mhd_model_file_path")
@click.argument("target_mhd_model_file_url")
def create_announcement_file_task(
    mhd_study_id: str,
    mhd_model_file_path: str,
    target_mhd_model_file_url: str,
    output_dir: str,
    output_filename: str,
):
    """Create announcement file from MHD data model file.

    Args:

    mhd_study_id (str): MHD study identifier

    mhd_model_file_path (str): MHD data model path

    target_mhd_model_file_url (str): target URL of MHD data model

    output_dir (str): Output directory of announcement file

    output_filename (str): Name of MHD announcement file. Default is <repository identifier>.announcement.json
    """
    file = Path(mhd_model_file_path)
    txt = file.read_text()
    mhd_data_json = json.loads(txt)

    Path(output_dir).mkdir(parents=True, exist_ok=True)

    announcement_file_path = f"{output_dir}/{mhd_study_id}.announcement.json"
    if output_filename:
        announcement_file_path = f"{output_dir}/{output_filename}"
    try:
        create_announcement_file(
            mhd_data_json, target_mhd_model_file_url, announcement_file_path
        )
        click.echo(f"{mhd_study_id} announcement file conversion completed.")
    except Exception as ex:
        click.echo(f"{mhd_study_id} announcement file conversion failed. {str(ex)}")



from mhd_model.shared.model import CvTerm
from mhd_model.shared.validation.definitions import ProfileValidationGroup
from mhd_model.shared.validation.registry import ProfileValidation


class CheckCvTermKeyValue(ProfileValidationGroup):
    name: str = "check-cv-term-key-value"
    cv_term_key: CvTerm
    min_value_count: int = 0


class CheckCvTermKeyValues(ProfileValidation):
    name: str = "check-cv-term-key-values"
    required_items: None | list[CheckCvTermKeyValue] = None
    optional_items: None | list[CheckCvTermKeyValue] = None
    input_type: str = "list"


class CheckChildCvTermKeyValues(ProfileValidation):
    name: str = "check-conditional-cv-term-key-values"

    conditional_field_name: str
    conditional_cv_term: CvTerm
    key_values_field_name: str
    key_values_control: CheckCvTermKeyValues



import json
import logging
from pathlib import Path
from typing import Any

from mhd_model.model.v0_1.dataset.profiles.legacy.profile import MhDatasetLegacyProfile

logger = logging.getLogger(__name__)


def create_neo4j_input_file(input_file_path: str, output_file_path: str):
    file = Path(input_file_path)
    txt = file.read_text()
    json_data = json.loads(txt)
    mhd_dataset = MhDatasetLegacyProfile.model_validate(json_data)
    nodes_map = {x.id_: x for x in mhd_dataset.graph.nodes}
    relationships_map = {x.id_: x for x in mhd_dataset.graph.relationships}
    nodes: list[dict[str, Any]] = []
    relationships: list[dict[str, Any]] = []
    for node in nodes_map.values():
        properties = {}
        for key, value in node.model_dump(exclude_none=True).items():
            refs = []
            if key == "id_":
                continue
            if key.endswith("_ref"):
                ref = getattr(node, key, None)
                if hasattr(nodes_map[ref], "name"):
                    val = getattr(nodes_map[ref], "name")
                    properties[key.replace("_ref", "")] = val
                refs.append(ref)
            elif key.endswith("_refs"):
                refs = getattr(node, key, [])
                vals = []
                for ref in refs:
                    if hasattr(nodes_map[ref], "name"):
                        val = getattr(nodes_map[ref], "name")
                        if val and val not in vals:
                            vals.append(val)
                properties[key.replace("_refs", "") + "_list"] = vals
            if refs:
                link_name = "EMBEDDED_" + key.replace("_ref", "").replace(
                    "_refs", ""
                ).upper().replace("-", "_")
                for ref in refs:
                    relationships.append(
                        {
                            "start": node.id_,
                            "end": ref,
                            "type": link_name,
                            "properties": {},
                        }
                    )

            if not refs:
                if isinstance(value, list):
                    properties[key] = [str(x) for x in value]
                elif isinstance(value, dict):
                    properties[key] = json.dumps({k: str(v) for k, v in value.items()})
                elif (
                    isinstance(value, str)
                    or isinstance(value, int)
                    or isinstance(value, float)
                    or isinstance(value, bool)
                ):
                    properties[key] = value
                else:
                    properties[key] = str(value)

        if not hasattr(properties, "name"):
            properties["name"] = node.label

        nodes.append({"id": node.id_, "labels": [node.type_], "properties": properties})

    for rel in relationships_map.values():
        relationships.append(
            {
                "start": rel.source_ref,
                "end": rel.target_ref,
                "type": rel.relationship_name.upper().replace("-", "_"),
                "properties": rel.model_dump(exclude_none=True),
            }
        )
    output_file = Path(output_file_path)
    with Path(output_file).open("w") as f:
        json.dump({"nodes": nodes, "relationships": relationships}, f, indent=2)



from pydantic import EmailStr, Field
from typing_extensions import Annotated

from mhd_model.model.v0_1.announcement.profiles.base.fields import (
    MeasurementType,
    MissingPublicationReason,
    OmicsType,
)
from mhd_model.model.v0_1.announcement.profiles.base.profile import (
    AnnouncementBaseProfile,
    AnnouncementContact,
    AnnouncementPublication,
    AnnouncementRawDataFile,
    AnnouncementResultFile,
)
from mhd_model.model.v0_1.announcement.profiles.ms import fields
from mhd_model.model.v0_1.dataset.profiles.base.base import (
    CvTerm,
)


class MsAnnouncementContact(AnnouncementContact):
    full_name: Annotated[str, Field(min_length=5)] = None
    emails: Annotated[list[EmailStr], Field(min_length=1)] = None
    affiliations: Annotated[list[str], Field(min_length=1)] = None


class AnnouncementMsProfile(AnnouncementBaseProfile):
    submitters: Annotated[list[MsAnnouncementContact], Field(min_length=1)]
    principal_investigators: Annotated[list[MsAnnouncementContact], Field(min_length=1)]

    # NMR, MS, ...
    technology_type: Annotated[list[fields.MsTechnologyType], Field(min_length=1)] = [
        CvTerm(
            source="OBI",
            accession="OBI:0000470",
            name="mass spectrometry assay",
        )
    ]
    # Targeted metabolite profiling, Untargeted metabolite profiling, ...
    measurement_type: Annotated[list[MeasurementType], Field(min_length=1)]
    # Metabolomics, Lipidomics, Proteomics, ...
    omics_type: Annotated[list[OmicsType], Field(min_length=1)]
    # LC-MS, GC-MS, ...
    assay_type: Annotated[list[fields.MsAssayType], Field(min_length=1)]

    publications: Annotated[
        MissingPublicationReason | list[AnnouncementPublication], Field()
    ]
    study_factors: Annotated[fields.StudyFactors, Field()]
    characteristic_values: Annotated[fields.ExtendedCharacteristicValues, Field()]
    protocols: Annotated[None | fields.Protocols, Field()] = None

    raw_data_file_list: Annotated[list[AnnouncementRawDataFile], Field()]
    result_file_list: Annotated[list[AnnouncementResultFile], Field(min_length=1)]
